import numpy as np
from bokeh import events

import xarray as xr
from arpes.plotting.interactive_utils import BokehInteractiveTool, CursorTool

__all__ = ('FitCheckTool',)


class FitCheckTool(BokehInteractiveTool, CursorTool):
    """
    Verification of fits
    """

    auto_zero_nans = False
    auto_rebin = False

    def __init__(self, **kwargs):
        super().__init__()

        self.load_settings(**kwargs)
        self.app_main_size = self.settings.get('main_width', 600)
        self.app_marginal_size = self.settings.get('marginal_width', 300)
        self.selected_data = 'data'
        self.use_dataset = True
        self.remove_outliers = True
        self.outlier_clip = 1

    def tool_handler(self, doc):
        from bokeh.layouts import row, column, widgetbox, Spacer
        from bokeh.models.mappers import LinearColorMapper
        from bokeh.models import widgets
        from bokeh.models.widgets.markups import Div
        from bokeh.plotting import figure

        self.arr = self.arr.copy(deep=True)

        if not isinstance(self.arr, xr.Dataset):
            self.use_dataset = False

        residual = None
        if self.use_dataset:
            raw_data = self.arr.data
            raw_data.values[np.isnan(raw_data.values)] = 0
            fit_results = self.arr.results
            residual = self.arr.residual
            residual.values[np.isnan(residual.values)] = 0
        else:
            raw_data = self.arr.attrs['original_data']
            fit_results = self.arr

        fit_direction = [d for d in raw_data.dims if d not in fit_results.dims]
        fit_direction = fit_direction[0]

        two_dimensional = False
        if len(raw_data.dims) != 2:
            two_dimensional = True
            x_coords, y_coords = fit_results.coords[fit_results.dims[0]], fit_results.coords[fit_results.dims[1]]
            z_coords = raw_data.coords[fit_direction]
        else:
            x_coords, y_coords = raw_data.coords[raw_data.dims[0]], raw_data.coords[raw_data.dims[1]]

        if two_dimensional:
            self.settings['palette'] = 'coolwarm'
        default_palette = self.default_palette

        self.app_context.update({
            'data': raw_data,
            'fits': fit_results,
            'residual': residual,
            'original': self.arr,
            'data_range': {
                'x': (np.min(x_coords.values), np.max(x_coords.values)),
                'y': (np.min(y_coords.values), np.max(y_coords.values)),
            }
        })
        if two_dimensional:
            self.app_context['data_range']['z'] = (np.min(z_coords.values), np.max(z_coords.values))

        figures, plots, app_widgets = self.app_context['figures'], self.app_context['plots'],\
                                      self.app_context['widgets']

        self.cursor_dims = raw_data.dims
        if two_dimensional:
            self.cursor = [np.mean(self.data_range['x']),
                           np.mean(self.data_range['y']),
                           np.mean(self.data_range['z'])]
        else:
            self.cursor = [np.mean(self.data_range['x']),
                           np.mean(self.data_range['y'])]

        app_widgets['fit_info_div'] = Div(text='')

        self.app_context['color_maps']['main'] = LinearColorMapper(
            default_palette, low=np.min(raw_data.values), high=np.max(raw_data.values), nan_color='black')

        main_tools = ["wheel_zoom", "tap", "reset","save"]
        main_title = 'Fit Inspection Tool: WARNING Unidentified'

        try:
            main_title = 'Fit Inspection Tool: {}'.format(raw_data.S.label[:60])
        except:
            pass

        figures['main'] = figure(
            tools=main_tools, plot_width=self.app_main_size, plot_height=self.app_main_size, min_border=10,
            min_border_left=50,
            toolbar_location='left', x_axis_location='below', y_axis_location='right',
            title=main_title, x_range=self.data_range['x'],
            y_range=self.app_context['data_range']['y'])
        figures['main'].xaxis.axis_label = raw_data.dims[0]
        figures['main'].yaxis.axis_label = raw_data.dims[1]
        figures['main'].toolbar.logo = None
        figures['main'].background_fill_color = "#fafafa"

        data_for_main = raw_data
        if two_dimensional:
            data_for_main = data_for_main.sel(**dict([[fit_direction, self.cursor[2]]]), method='nearest')
        plots['main'] = figures['main'].image(
            [data_for_main.values.T], x=self.app_context['data_range']['x'][0], y=self.app_context['data_range']['y'][0],
            dw=self.app_context['data_range']['x'][1] - self.app_context['data_range']['x'][0],
            dh=self.app_context['data_range']['y'][1] - self.app_context['data_range']['y'][0],
            color_mapper=self.app_context['color_maps']['main'])

        band_centers = [b.center for b in fit_results.F.bands.values()]
        bands_xs = [b.coords[b.dims[0]].values for b in band_centers]
        bands_ys = [b.values for b in band_centers]
        if fit_results.dims[0] == raw_data.dims[1]:
            bands_ys, bands_xs = bands_xs, bands_ys
        plots['band_locations'] = figures['main'].multi_line(
            xs=bands_xs, ys=bands_ys,
            line_color='white', line_width=1, line_dash='dashed')

        # add cursor lines
        cursor_lines = self.add_cursor_lines(figures['main'])

        # marginals
        if not two_dimensional:
            figures['bottom'] = figure(
                plot_width=self.app_main_size, plot_height=self.app_marginal_size, min_border=10,
                title=None, x_range=figures['main'].x_range,
                x_axis_location='above',
                toolbar_location=None, tools=[])
        else:
            figures['bottom'] = Spacer(width=self.app_main_size, height=self.app_marginal_size)

        right_y_range = figures['main'].y_range
        if two_dimensional:
            right_y_range = self.data_range['z']

        figures['right'] = figure(
            plot_width=self.app_marginal_size, plot_height=self.app_main_size, min_border=10,
            title=None, y_range=right_y_range,
            y_axis_location='left',
            toolbar_location=None, tools=[])

        marginal_line_width = 2
        if not two_dimensional:
            bottom_data = raw_data.sel(**dict([[raw_data.dims[1], self.cursor[1]]]), method='nearest')
            right_data = raw_data.sel(**dict([[raw_data.dims[0], self.cursor[0]]]), method='nearest')

            plots['bottom'] = figures['bottom'].line(x=bottom_data.coords[raw_data.dims[0]].values,
                                                     y=bottom_data.values, line_width=marginal_line_width)
            plots['bottom_residual'] = figures['bottom'].line(x=[], y=[], line_color='red',
                                                              line_width=marginal_line_width)
            plots['bottom_fit'] = figures['bottom'].line(
                x=[], y=[], line_color='blue', line_width=marginal_line_width, line_dash='dashed')
            plots['bottom_init_fit'] = figures['bottom'].line(
                x=[], y=[], line_color='green', line_width=marginal_line_width, line_dash='dotted')

            plots['right'] = figures['right'].line(y=right_data.coords[raw_data.dims[1]].values, x=right_data.values,
                                                   line_width=marginal_line_width)
            plots['right_residual'] = figures['right'].line(x=[], y=[], line_color='red',
                                                            line_width=marginal_line_width)
            plots['right_fit'] = figures['right'].line(
                x=[], y=[], line_color='blue', line_width=marginal_line_width, line_dash='dashed')
            plots['right_init_fit'] = figures['right'].line(
                x=[], y=[], line_color='green', line_width=marginal_line_width, line_dash='dotted')
        else:
            right_data = raw_data.sel(**{k: v for k, v in self.cursor_dict.items() if k != fit_direction}, method='nearest')
            plots['right'] = figures['right'].line(y=right_data.coords[right_data.dims[0]].values, x=right_data.values,
                                                   line_width=marginal_line_width)
            plots['right_residual'] = figures['right'].line(x=[], y=[], line_color='red',
                                                            line_width=marginal_line_width)
            plots['right_fit'] = figures['right'].line(
                x=[], y=[], line_color='blue', line_width=marginal_line_width, line_dash='dashed')
            plots['right_init_fit'] = figures['right'].line(
                x=[], y=[], line_color='green', line_width=marginal_line_width, line_dash='dotted')

        def on_change_main_view(attr, old, data_source):
            self.selected_data = data_source
            data = None
            if data_source == 'data':
                data = raw_data.sel(**{k: v for k, v in self.cursor_dict.items() if k == fit_direction},
                                    method='nearest')
            elif data_source == 'residual':
                data = residual.sel(**{k: v for k, v in self.cursor_dict.items() if k == fit_direction},
                                    method='nearest')
            elif two_dimensional:
                data = fit_results.F.s(data_source)
                data.values[np.isnan(data.values)] = 0

            if data is not None:
                if self.remove_outliers:
                    data = data.T.clean_outliers(clip=self.outlier_clip)

                plots['main'].data_source.data = {
                    'image': [data.values.T],
                }
                update_main_colormap(None, None, main_color_range_slider.value)

        def update_fit_display():
            target = 'right'
            if fit_results.dims[0] == raw_data.dims[1]:
                target = 'bottom'

            if two_dimensional:
                target = 'right'
                current_fit = fit_results.sel(**{k: v for k, v in self.cursor_dict.items() if k != fit_direction}, method='nearest').item()
                coord_vals = raw_data.coords[fit_direction].values
            else:
                current_fit = fit_results.sel(**dict([[fit_results.dims[0], self.cursor[0 if target == 'right' else 1]]]), method='nearest').item()
                coord_vals = raw_data.coords[raw_data.dims[0 if target == 'bottom' else 1]].values

            if current_fit is not None:
                app_widgets['fit_info_div'].text = current_fit._repr_html_(short=True) # pylint: disable=protected-access
            else:
                app_widgets['fit_info_div'].text = 'No fit here.'
                plots['{}_residual'.format(target)].data_source.data = {
                    'x': [], 'y': [],
                }
                plots['{}_fit'.format(target)].data_source.data = {
                    'x': [], 'y': [],
                }
                plots['{}_init_fit'.format(target)].data_source.data = {
                    'x': [], 'y': [],
                }
                return

            if target == 'bottom':
                residual_x = coord_vals
                residual_y = current_fit.residual
                init_fit_x = coord_vals
                init_fit_y = current_fit.init_fit
                fit_x = coord_vals
                fit_y = current_fit.best_fit
            else:
                residual_y = coord_vals
                residual_x = current_fit.residual
                init_fit_y = coord_vals
                init_fit_x = current_fit.init_fit
                fit_y = coord_vals
                fit_x = current_fit.best_fit

            plots['{}_residual'.format(target)].data_source.data = {
                'x': residual_x,
                'y': residual_y,
            }
            plots['{}_fit'.format(target)].data_source.data = {
                'x': fit_x,
                'y': fit_y,
            }
            plots['{}_init_fit'.format(target)].data_source.data = {
                'x': init_fit_x,
                'y': init_fit_y,
            }

        def click_right_marginal(event):
            self.cursor = [self.cursor[0], self.cursor[1], event.y]
            on_change_main_view(None, None, self.selected_data)

        def click_main_image(event):
            if two_dimensional:
                self.cursor = [event.x, event.y, self.cursor[2]]
            else:
                self.cursor = [event.x, event.y]

            if not two_dimensional:
                right_marginal_data = raw_data.sel(**dict([[raw_data.dims[0], self.cursor[0]]]), method='nearest')
                bottom_marginal_data = raw_data.sel(**dict([[raw_data.dims[1], self.cursor[1]]]), method='nearest')
                plots['bottom'].data_source.data = {
                    'x': bottom_marginal_data.coords[raw_data.dims[0]].values,
                    'y': bottom_marginal_data.values,
                }
            else:
                right_marginal_data = raw_data.sel(**{k: v for k, v in self.cursor_dict.items() if k != fit_direction}, method='nearest')

            plots['right'].data_source.data = {
                'y': right_marginal_data.coords[right_marginal_data.dims[0]].values,
                'x': right_marginal_data.values,
            }

            update_fit_display()

        def on_change_outlier_clip(attr, old, new):
            self.outlier_clip = new
            on_change_main_view(None, None, self.selected_data)

        def set_remove_outliers(should_remove_outliers):
            if self.remove_outliers != should_remove_outliers:
                self.remove_outliers = should_remove_outliers

                on_change_main_view(None, None, self.selected_data)

        update_main_colormap = self.update_colormap_for('main')
        MAIN_CONTENT_OPTIONS = [
            ('Residual', 'residual'),
            ('Data', 'data'),
        ]

        if two_dimensional:
            available_parameters = fit_results.F.parameter_names

            for param_name in available_parameters:
                MAIN_CONTENT_OPTIONS.append((param_name, param_name,))

        remove_outliers_toggle = widgets.Toggle(label='Remove Outliers', button_type='primary', active=self.remove_outliers)
        remove_outliers_toggle.on_click(set_remove_outliers)

        outlier_clip_slider = widgets.Slider(title='Clip', start=0, end=10, value=self.outlier_clip,
                                             callback_throttle=150, step=0.2)
        outlier_clip_slider.on_change('value', on_change_outlier_clip)

        main_content_select = widgets.Dropdown(label='Main Content', button_type='primary', menu=MAIN_CONTENT_OPTIONS)
        main_content_select.on_change('value', on_change_main_view)

        # Widgety things
        main_color_range_slider = widgets.RangeSlider(
            start=0, end=100, value=(0, 100,), title='Color Range')

        # Attach callbacks
        main_color_range_slider.on_change('value', update_main_colormap)
        figures['main'].on_event(events.Tap, click_main_image)
        if two_dimensional:
            figures['right'].on_event(events.Tap, click_right_marginal)

        layout = row(column(figures['main'], figures.get('bottom')),
                     column(figures['right'], app_widgets['fit_info_div']),
                     column(
                         widgetbox(
                             *[widget for widget in [
                                 self._cursor_info,
                                 main_color_range_slider,
                                 main_content_select,
                                 remove_outliers_toggle if two_dimensional else None,
                                 outlier_clip_slider if two_dimensional else None,
                             ] if widget is not None]
                         ),
                     ))

        update_fit_display()

        doc.add_root(layout)
        doc.title = 'Band Tool'



import numpy as np
from scipy.spatial import ConvexHull

__all__ = ('polyhedron_intersect_plane', 'segment_contains_point', 'point_plane_intersection',)


def point_plane_intersection(plane_normal, plane_point, line_a, line_b, epsilon=1e-6):
    """
    Determines the intersection point of a plane defined by a point and a normal vector and the line
    defined by line_a and line_b. All should be numpy arrays.

    :param plane_normal:
    :param plane_point:
    :param line_a:
    :param line_b:
    :param epsilon:
    :return:
    """

    line_direction = line_b - line_a
    if abs(plane_normal.dot(line_direction)) < epsilon:
        return None

    delta = line_a - plane_point
    projection = -plane_normal.dot(delta) / plane_normal.dot(line_direction)

    return delta + projection * line_direction + plane_point


def segment_contains_point(line_a, line_b, point_along_line, check=False, epsilon=1e-6):
    """
    Determines whether a segment contains a point that also lies along the line.
    If asked to check, it will also return false if the point does not lie along the line.

    :param line_a:
    :param line_b:
    :param point_along_line:
    :param check:
    :param epsilon:
    :return:
    """
    if point_along_line is None:
        return False

    delta = line_b - line_a
    delta_p = point_along_line - line_a
    if check:
        cosine = delta.dot(delta_p) / (np.linalg.norm(delta) * np.linalg.norm(delta_p))
        if cosine < 1 - epsilon:
            return False

    return 0 - epsilon < delta.dot(delta_p) / delta.dot(delta) < 1 + epsilon


def polyhedron_intersect_plane(poly_faces, plane_normal, plane_point, epsilon=1e-6):
    """
    Determines the intersection of a convex polyhedron intersecting a plane. The polyhedron
    faces should be given by a list of np.arrays, where each np.array at index `i` is the vertices of
    face `i`.

    As an example, running [p[0] for p in ase.dft.bz.bz_vertices(np.linalg.inv(cell).T)]
    should provide acceptable input for a unit cell `cell`.

    The polyhedron should be convex because we construct the convex hull in order to order the points.

    :param poly_faces:
    :param plane_normal:
    :param plane_point:
    :param epsilon:
    :return:
    """

    collected_points = []

    def add_point(c):
        already_collected = False
        for other in collected_points:
            delta = c - other
            if delta.dot(delta) < epsilon:
                already_collected = True
                break

        if not already_collected:
            collected_points.append(c)

    for poly_face in poly_faces:
        segments = list(zip(poly_face, np.concatenate([poly_face[1:], [poly_face[0]]])))
        for a, b in segments:
            intersection = point_plane_intersection(plane_normal, plane_point, a, b, epsilon=epsilon)
            if segment_contains_point(a, b, intersection, epsilon=epsilon):
                add_point(intersection)

    points = ConvexHull(collected_points, qhull_options='Qc QJ').points

    # sort
    for_sort = points - np.mean(points, axis=0)
    for_sort = (for_sort.T / np.linalg.norm(for_sort, axis=1)).T
    det = plane_normal.dot(np.cross(for_sort[0], for_sort).T)
    dot = for_sort[0].dot(for_sort.T)

    return points[np.argsort(np.arctan2(det, dot))]



import numpy as np
import scipy
import scipy.ndimage
from tqdm import tqdm_notebook

import xarray as xr
from arpes.fits.fit_models import gaussian
from arpes.provenance import update_provenance
from arpes.typing import DataType
from arpes.utilities import normalize_to_spectrum

__all__ = ('deconvolve_ice', 'deconvolve_rl', 'make_psf1d',)


@update_provenance('Approximate Iterative Deconvolution')
def deconvolve_ice(data: DataType, psf, n_iterations=5, deg=None):
    """
    Deconvolves data by a given point spread function using the iterative convolution extrapolation method.
    :param data:
    :param psf:
    :param n_iterations -- the number of convolutions to use for the fit (default 5):
    :param deg -- the degree of the fitting polynominal (default n_iterations-3):
    :return DataArray or numpy.ndarray -- based on input type:
    """

    arr = normalize_to_spectrum(data)
    if type(data) is np.ndarray:
        pass
    else:
        arr = arr.values

    if deg is None:
        deg = n_iterations - 3
    iteration_steps = list(range(1, n_iterations + 1))

    iteration_list = [arr]

    for i in range(n_iterations - 1):
        iteration_list.append(scipy.ndimage.convolve(iteration_list[-1], psf))
    iteration_list = np.asarray(iteration_list)

    deconv = arr * 0
    for t, series in enumerate(iteration_list.T):
        coefs = np.polyfit(iteration_steps, series, deg=deg)
        poly = np.poly1d(coefs)
        deconv[t] = poly(0)

    if type(data) is np.ndarray:
        result = deconv
    else:
        result = normalize_to_spectrum(data).copy(deep=True)
        result.values = deconv
    return result


@update_provenance('Lucy Richardson Deconvolution')
def deconvolve_rl(data: DataType, psf=None, n_iterations=10, axis=None,
                  sigma=None, mode='reflect', progress=True):
    """Deconvolves data by a given point spread function using the Richardson-Lucy method.

    :param data:
    :param psf -- for 1d, if not specified, must specify axis and sigma:
    :param n_iterations -- the number of convolutions to use for the fit (default 50):
    :param axis:
    :param sigma:
    :param mode:
    :param progress:
    :return DataArray or numpy.ndarray -- based on input type:
    """

    arr = normalize_to_spectrum(data)

    if psf is None and axis is not None and sigma is not None:
        # if no psf is provided and we have the information to make a 1d one
        # note: this assumes gaussian psf
        psf = make_psf1d(data=arr, dim=axis, sigma=sigma)

    if len(data.dims) > 1:
        if axis is not None:
            # perform one-dimensional deconvolution of multidimensional data

            # support for progress bars
            wrap_progress = lambda x, *args, **kwargs: x
            if progress:
                wrap_progress = tqdm_notebook

            # dimensions over which to iterate
            other_dim = list(data.dims)
            other_dim.remove(axis)

            if len(other_dim) == 1:
                # two-dimensional data
                other_dim = other_dim[0]
                result = arr.copy(deep=True).transpose(other_dim,
                                                       axis)  # not sure why the dims only seems to work in this order. seems like I should be able to swap it to (axis,other_dim) and also change the data collection to result[x_ind,y_ind], but this gave different results

                for i, (od, iteration) in wrap_progress(enumerate(arr.T.iterate_axis(other_dim)),
                                                        desc="Iterating " + other_dim,
                                                        total=len(arr[other_dim])):  # TODO tidy this gross-looking loop
                    # indices of data being deconvolved
                    x_ind = xr.DataArray(list(range(len(arr[axis]))), dims=[axis])
                    y_ind = xr.DataArray([i] * len(x_ind), dims=[other_dim])
                    # perform deconvolution on this one-dimensional piece
                    deconv = deconvolve_rl(data=iteration, psf=psf, n_iterations=n_iterations, axis=None, mode=mode)
                    # build results out of these pieces
                    result[y_ind, x_ind] = deconv.values
            elif len(other_dim) == 2:
                # three-dimensional data
                result = arr.copy(deep=True).transpose(*other_dim,
                                                       axis)  # not sure why the dims only seems to work in this order. seems like I should be able to swap it to (axis,*other_dim) and also change the data collection to result[x_ind,y_ind,z_ind], but this gave different results
                for i, (od0, iteration0) in wrap_progress(enumerate(arr.T.iterate_axis(other_dim[0])),
                                                          desc="Iterating " + other_dim[0], total=len(
                                arr[other_dim[0]])):  # TODO tidy this gross-looking loop
                    for j, (od1, iteration1) in wrap_progress(enumerate(iteration0.T.iterate_axis(other_dim[1])),
                                                              desc="Iterating " + other_dim[1],
                                                              total=len(arr[other_dim[1]]),
                                                              leave=False):  # TODO tidy this gross-looking loop
                        # indices of data being deconvolved
                        x_ind = xr.DataArray(list(range(len(arr[axis]))), dims=[axis])
                        y_ind = xr.DataArray([i] * len(x_ind), dims=[other_dim[0]])
                        z_ind = xr.DataArray([j] * len(x_ind), dims=[other_dim[1]])
                        # perform deconvolution on this one-dimensional piece
                        deconv = deconvolve_rl(data=iteration1, psf=psf, n_iterations=n_iterations, axis=None,
                                               mode=mode)
                        # build results out of these pieces
                        result[y_ind, z_ind, x_ind] = deconv.values
            elif len(other_dim) >= 3:
                # four- or higher-dimensional data
                # TODO find way to compactify the different dimensionalities rather than having separate code
                raise NotImplementedError("high-dimensional data not yet supported")

        elif axis is None:
            # crude attempt to perform multidimensional deconvolution. not clear if this is currently working
            # TODO may be able to do this as a sequence of one-dimensional deconvolutions, assuming that the psf is separable (which I think it should be, if we assume it is a multivariate gaussian with principle axes aligned with the dimensions)
            raise NotImplementedError("multi-dimensional convolutions not yet supported")

            if not isinstance(arr, np.ndarray):
                arr = arr.values

            u = [arr]

            for i in range(n_iterations):
                c = scipy.ndimage.convolve(u[-1], psf, mode=mode)
                u.append(u[-1] * scipy.ndimage.convolve(arr / c, np.flip(psf, None),
                                                        mode=mode))  # careful about which axis (axes) to flip here...! need to explicitly specify for some versions of numpy

            result = u[-1]
    else:
        if type(arr) is not np.ndarray:
            arr = arr.values

        u = [arr]

        for i in range(n_iterations):
            c = scipy.ndimage.convolve(u[-1], psf, mode=mode)
            u.append(u[-1] * scipy.ndimage.convolve(arr / c, np.flip(psf, 0),
                                                    mode=mode))  # not yet tested to ensure flip correct for asymmetric psf
            # note: need to explicitly specify axis number in np.flip in lower versions of numpy

        if type(data) is np.ndarray:
            result = u[-1].copy()
        else:
            result = normalize_to_spectrum(data).copy(deep=True)
            result.values = u[-1]

    try:
        result = result.transpose(*arr.dims)
    except:
        pass

    return result


@update_provenance('Make 1D-Point Spread Function')
def make_psf1d(data: DataType, dim, sigma):
    """Produces a 1-dimensional gaussian point spread function for use in deconvolve_rl.

    :param data:
    :param dim:
    :param sigma:
    :return DataArray:
    """

    arr = normalize_to_spectrum(data)
    dims = arr.dims

    psf = arr.copy(deep=True) * 0 + 1

    other_dims = list(arr.dims)
    other_dims.remove(dim)

    for od in other_dims:
        psf = psf[{od: 0}]

    psf = psf * gaussian(psf.coords[dim], np.mean(psf.coords[dim]), sigma)

    return psf


@update_provenance('Make Point Spread Function')
def make_psf(data: DataType, sigmas):
    """Not yet operational; produces an n-dimensional gaussian point spread function for use in deconvolve_rl.

    :param data:
    :param dim:
    :param sigma:
    :return DataArray:
    """

    raise NotImplementedError()

    arr = normalize_to_spectrum(data)
    dims = arr.dims

    psf = arr.copy(deep=True) * 0 + 1

    for dim in dims:
        other_dims = list(arr.dims)
        other_dims.remove(dim)

        psf1d = arr.copy(deep=True) * 0 + 1
        for od in other_dims:
            psf1d = psf1d[{od: 0}]

        if sigmas[dim] == 0:
            # TODO may need to do subpixel correction for when the dimension has an even length
            psf1d = psf1d * 0
            # psf1d[{dim:np.mean(psf1d.coords[dim])}] = 1
            psf1d[{dim: len(psf1d.coords[dim]) / 2}] = 1
        else:
            psf1d = psf1d * gaussian(psf1d.coords[dim], np.mean(psf1d.coords[dim]), sigmas[dim])

        psf = psf * psf1d
    return psf



import xarray as xr
from arpes.analysis import savitzky_golay
from arpes.provenance import update_provenance
from arpes.typing import DataType
from arpes.utilities import normalize_to_dataset
from arpes.utilities.math import polarization

__all__ = ('to_intensity_polarization', 'to_up_down', 'normalize_sarpes_photocurrent', 'sarpes_smooth')


@update_provenance('Smooth SARPES data')
def sarpes_smooth(data: xr.Dataset, *args, **kwargs):
    """
    Smooths the up and down channels.
    :param data:
    :param args:
    :param kwargs:
    :return:
    """
    up = savitzky_golay(data.up, *args, **kwargs)
    down = savitzky_golay(data.down, *args, **kwargs)
    return data.copy(deep=True).assign(up=up, down=down)


@update_provenance('Normalize SARPES by photocurrent')
def normalize_sarpes_photocurrent(data: DataType):
    """
    Normalizes the down channel so that it matches the up channel in terms of mean photocurrent. Destroys the integrity
    of "count" data.

    :param data:
    :return:
    """

    copied = data.copy(deep=True)
    copied.down.values = (copied.down * (copied.photocurrent_up / copied.photocurrent_down)).values
    return copied


@update_provenance('Convert polarization data to up-down spin channels')
def to_up_down(data: DataType):
    """
    Converts from [intensity, polarization] representation to [up, down] representation.
    :param data:
    :return:
    """
    assert('intensity' in data.data_vars and 'polarization' in data.data_vars)

    return xr.Dataset({
        'up': data.intensity * (1 + data.polarization),
        'down': data.intensity * (1 - data.polarization),
    })


@update_provenance('Convert up-down spin channels to polarization')
def to_intensity_polarization(data: DataType):
    """
    Converts from [up, down] representation (the spin projection) to
    [intensity, polarization] representation.

    In this future, we should also make this also work with the timing signals.
    :param data:
    :return:
    """
    data = normalize_to_dataset(data)

    assert('up' in data.data_vars and 'down' in data.data_vars)

    intensity = data.up + data.down
    spectrum_polarization = polarization(data.up, data.down)

    return xr.Dataset({
        'intensity': intensity,
        'polarization': spectrum_polarization / data.S.sherman_function
    })



import numpy as np

from arpes.analysis.mask import apply_mask
from arpes.exceptions import AnalysisError
from arpes.plotting.interactive_utils import CursorTool, SaveableTool
from arpes.typing import DataType
from arpes.utilities import normalize_to_spectrum

__all__ = ('MaskTool', 'mask')


class MaskTool(SaveableTool, CursorTool):
    """
    Tool to allow masking data by drawing regions.
    """

    auto_zero_nans = False
    auto_rebin = False

    def __init__(self, **kwargs):
        super().__init__(kwargs.pop('name', None))

        self.load_settings(**kwargs)

        self.app_main_size = self.settings.get('app_main_size', 600)
        self.app_marginal_size = self.settings.get('app_main_size', 300)
        self.pointer_mode = 'region'

    def tool_handler(self, doc):
        from bokeh import events
        from bokeh.layouts import row, column, widgetbox
        from bokeh.models.mappers import LinearColorMapper
        from bokeh.models import widgets
        from bokeh.plotting import figure

        if len(self.arr.shape) != 2:
            raise AnalysisError('Cannot use mask tool on non image-like spectra')

        arr = self.arr
        x_coords, y_coords = arr.coords[arr.dims[0]], arr.coords[arr.dims[1]]

        default_palette = self.default_palette

        self.app_context.update({
            'region_options': [],
            'regions': {},
            'data': arr,
            'data_range': {
                'x': (np.min(x_coords.values), np.max(x_coords.values)),
                'y': (np.min(y_coords.values), np.max(y_coords.values)),
            },
        })

        self.cursor = [np.mean(self.data_range['x']),
                       np.mean(self.data_range['y'])]

        self.color_maps['main'] = LinearColorMapper(default_palette, low=np.min(arr.values), high=np.max(arr.values),
                                                    nan_color='black')

        main_tools = ["wheel_zoom", "tap", "reset"]
        main_title = 'Mask Tool: WARNING Unidentified'

        try:
            main_title = 'Mask Tool: {}'.format(arr.S.label[:60])
        except:
            pass

        self.figures['main'] = figure(
            tools=main_tools, plot_width=self.app_main_size, plot_height=self.app_main_size, min_border=10,
            min_border_left=20, toolbar_location='left', x_axis_location='below', y_axis_location='right',
            title=main_title, x_range=self.data_range['x'], y_range=self.data_range['y'],
        )

        self.figures['main'].xaxis.axis_label = arr.dims[0]
        self.figures['main'].yaxis.axis_label = arr.dims[1]

        self.plots['main'] = self.figures['main'].image(
            [np.asarray(arr.values.T)], x=self.data_range['x'][0], y=self.data_range['y'][0],
            dw=self.data_range['x'][1] - self.data_range['x'][0],
            dh=self.data_range['y'][1] - self.data_range['y'][0],
            color_mapper=self.color_maps['main']
        )

        self.add_cursor_lines(self.figures['main'])
        region_patches = self.figures['main'].patches(xs=[], ys=[], color='white', alpha=0.35, line_width=1)

        def add_point_to_region():
            if self.active_region in self.regions:
                self.regions[self.active_region]['points'].append(list(self.cursor))
                update_region_display()

            self.save_app()

        def click_main_image(event):
            self.cursor = [event.x, event.y]
            if self.pointer_mode == 'region':
                add_point_to_region()

        POINTER_MODES = [
            ('Cursor', 'cursor',),
            ('Region', 'region',),
        ]

        def perform_mask(data=None, **kwargs):
            if data is None:
                data = arr

            data = normalize_to_spectrum(data)
            return apply_mask(data, self.app_context['mask'], **kwargs)

        self.app_context['perform_mask'] = perform_mask
        self.app_context['mask'] = None

        pointer_dropdown = widgets.Dropdown(label='Pointer Mode', button_type='primary', menu=POINTER_MODES)
        self.region_dropdown = widgets.Dropdown(label='Active Region', button_type='primary',
                                                menu=self.region_options)

        edge_mask_button = widgets.Button(label='Edge Mask')
        region_name_input = widgets.TextInput(placeholder='Region name...')
        add_region_button = widgets.Button(label='Add Region')

        clear_region_button = widgets.Button(label='Clear Region')
        remove_region_button = widgets.Button(label='Remove Region')

        main_color_range_slider = widgets.RangeSlider(
            start=0, end=100, value=(0, 100,), title='Color Range')

        def on_click_edge_mask():
            if self.active_region in self.regions:
                old_points = self.regions[self.active_region]['points']
                dims = [d for d in arr.dims if 'eV' != d]
                energy_index = arr.dims.index('eV')
                max_energy = np.max([p[energy_index] for p in old_points])

                other_dim = dims[0]
                other_coord = arr.coords[other_dim].values
                min_other, max_other = np.min(other_coord), np.max(other_coord)
                min_e = np.min(arr.coords['eV'].values)

                if arr.dims.index('eV') == 0:
                    before = [[min_e - 3, min_other - 1], [0, min_other - 1]]
                    after = [[0, max_other + 1], [min_e - 3, max_other + 1]]
                else:
                    before = [[min_other - 1, min_e - 3], [min_other - 1, 0]]
                    after = [[max_other + 1, 0], [max_other + 1, min_e - 3]]
                self.regions[self.active_region]['points'] = before + old_points + after
                self.app_context['mask'] = self.app_context['mask'] or {}
                self.app_context['mask']['fermi'] = max_energy
                update_region_display()

            self.save_app()

        def add_region(region_name):
            if region_name not in self.regions:
                self.region_options.append((region_name, region_name,))
                self.region_dropdown.menu = self.region_options
                self.regions[region_name] = {
                    'points': [],
                    'name': region_name,
                }

                if self.active_region is None:
                    self.active_region = region_name

                self.save_app()

        def on_change_active_region(attr, old_region_id, region_id):
            self.app_context['active_region'] = region_id
            self.active_region = region_id
            self.save_app()

        def on_change_pointer_mode(attr, old_pointer_mode, pointer_mode):
            self.app_context['pointer_mode'] = pointer_mode
            self.pointer_mode = pointer_mode
            self.save_app()

        def update_region_display():
            region_names = self.regions.keys()

            if self.app_context['mask'] is None:
                self.app_context['mask'] = {}
            self.app_context['mask'].update({
                'dims': arr.dims,
                'polys': [r['points'] for r in self.regions.values()]
            })

            region_patches.data_source.data = {
                'xs': [[p[0] for p in self.regions[r]['points']] for r in region_names],
                'ys': [[p[1] for p in self.regions[r]['points']] for r in region_names],
            }
            self.save_app()

        self.update_region_display = update_region_display

        def on_clear_region():
            if self.active_region in self.regions:
                self.regions[self.active_region]['points'] = []
                update_region_display()

        def on_remove_region():
            if self.active_region in self.regions:
                del self.regions[self.active_region]
                new_region_options = [b for b in self.region_options if b[0] != self.active_region]
                self.region_dropdown.menu = new_region_options
                self.region_options = new_region_options
                self.active_region = None
                update_region_display()

        # Attach callbacks
        main_color_range_slider.on_change('value', self.update_colormap_for('main'))

        self.figures['main'].on_event(events.Tap, click_main_image)
        self.region_dropdown.on_change('value', on_change_active_region)
        pointer_dropdown.on_change('value', on_change_pointer_mode)
        add_region_button.on_click(lambda: add_region(region_name_input.value))
        edge_mask_button.on_click(on_click_edge_mask)
        clear_region_button.on_click(on_clear_region)
        remove_region_button.on_click(on_remove_region)

        layout = row(column(self.figures['main']),
                     column(
                         *[f for f in [
                             widgetbox(
                                 pointer_dropdown,
                                 self.region_dropdown,
                             ),
                             row(
                                 region_name_input,
                                 add_region_button,
                             ),
                             edge_mask_button if 'eV' in arr.dims else None,
                             row(
                                 clear_region_button,
                                 remove_region_button,
                             ),
                             widgetbox(
                                 self._cursor_info,
                                 main_color_range_slider,
                             ),
                         ] if f is not None]
                     ))

        doc.add_root(layout)
        doc.title = 'Mask Tool'
        self.load_app()
        self.save_app()

    def serialize(self):
        return {
            'active_region': self.active_region,
            'region_options': self.region_options,
            'regions': self.regions,
            'cursor': self.cursor,
        }

    def deserialize(self, json_data):
        self.cursor = json_data.get('cursor', [0, 0])

        self.app_context['regions'] = json_data.get('regions', {}) or {}
        self.app_context['region_options'] = json_data.get('region_options', [])

        self.region_dropdown.menu = self.app_context['region_options']

        self.update_region_display()


def mask(data: DataType, **kwargs):
    data = normalize_to_spectrum(data)

    tool = MaskTool(**kwargs)
    return tool.make_tool(data)



"""
Helper functions for coordinate transformations. All the functions here
assume standard polar angles, as given in the
`data model documentation <https://arpes.netlify.com/#/spectra>`_.


Functions here must accept constants or numpy arrays as valid inputs,
so all standard math functions have been replaced by their equivalents out
of numpy. Array broadcasting should handle any issues or weirdnesses that
would encourage the use of direct iteration, but in case you need to write
a conversion directly, be aware that any functions here must work on arrays
as well for consistency with client code.

Everywhere:

Kinetic energy -> 'kinetic_energy'
Binding energy -> 'eV', for convenience (negative below 0)
Photon energy -> 'hv'

Better facilities should be added for ToFs to do simultaneous (timing, angle) to (binding energy, k-space).
"""

import collections
import warnings
from copy import deepcopy

import numpy as np
import scipy.interpolate

import xarray as xr
from arpes.exceptions import AnalysisError
from arpes.provenance import provenance, update_provenance
from arpes.utilities import normalize_to_spectrum
from typing import Union

from .kx_ky_conversion import ConvertKxKy, ConvertKp
from .kz_conversion import ConvertKpKz

__all__ = ['convert_to_kspace', 'slice_along_path']


def infer_kspace_coordinate_transform(arr: xr.DataArray):
    """
    Infers appropriate coordinate transform for arr to momentum space.

    This takes into account the extra metadata attached to arr that might be
    useful in inferring the requirements of the coordinate transform, like the
    orientation of the spectrometer slit, and other experimental concerns
    :param arr:
    :return: dict with keys ``target_coordinates``, and a map of the appropriate
    conversion functions
    """
    old_coords = deepcopy(list(arr.coords))
    assert 'eV' in old_coords
    old_coords.remove('eV')
    old_coords.sort()

    new_coords = {
        ('phi',): ['kp'],
        ('beta', 'phi',): ['kx', 'ky'],
        ('phi', 'theta',): ['kx', 'ky'],
        ('phi', 'psi',): ['kx', 'ky'],
        ('hv', 'phi',): ['kp', 'kz'],
        ('beta', 'hv', 'phi',): ['kx', 'ky', 'kz'],
        ('hv', 'phi', 'theta',): ['kx', 'ky', 'kz'],
        ('hv', 'phi', 'psi',): ['kx', 'ky', 'kz'],
    }.get(tuple(old_coords))

    # At this point we need to do a bit of work in order to determine the functions
    # that interpolate from k-space back into the recorded variable space

    # TODO Also provide the Jacobian of the coordinate transform to properly
    return {
        'dims': new_coords,
        'transforms': {

        },
        'calculate_bounds': None,
        'jacobian': None,
    }


def grid_interpolator_from_dataarray(arr: xr.DataArray, fill_value=0.0, method='linear',
                                     bounds_error=False):
    """
    Translates the contents of an xarray.DataArray into a scipy.interpolate.RegularGridInterpolator.

    This is principally used for coordinate translations.
    """
    flip_axes = set()
    for d in arr.dims:
        c = arr.coords[d]
        if len(c) > 1 and c[1] - c[0] < 0:
            flip_axes.add(d)

    values = arr.values
    for dim in flip_axes:
        values = np.flip(values, arr.dims.index(dim))

    return scipy.interpolate.RegularGridInterpolator(
        points=[arr.coords[d].values[::-1] if d in flip_axes else arr.coords[d].values for d in arr.dims],
        values=values,
        bounds_error=bounds_error, fill_value=fill_value, method=method)


def slice_along_path(arr: xr.DataArray, interpolation_points=None, axis_name=None, resolution=None,
                     shift_gamma=True, extend_to_edge=False, **kwargs):
    """
    TODO: There might be a little bug here where the last coordinate has a value of 0, causing the interpolation to loop
    back to the start point. For now I will just deal with this in client code where I see it until I understand if it is
    universal.

    Interpolates along a path through a volume. If the volume is higher dimensional than the desired path, the
    interpolation is broadcasted along the free dimensions. This allows one to specify a k-space path and receive
    the band structure along this path in k-space.

    Points can either by specified by coordinates, or by reference to symmetry points, should they exist in the source
    array. These symmetry points are translated to regular coordinates immediately, but are provided as a convenience.
    If not all points specify the same set of coordinates, an attempt will be made to unify the coordinates. As an example,
    if the specified path is (kx=0, ky=0, T=20) -> (kx=1, ky=1), the path will be made between (kx=0, ky=0, T=20) ->
    (kx=1, ky=1, T=20). On the other hand, the path (kx=0, ky=0, T=20) -> (kx=1, ky=1, T=40) -> (kx=0, ky=1) will result
    in an error because there is no way to break the ambiguity on the temperature for the last coordinate.

    A reasonable value will be chosen for the resolution, near the maximum resolution of any of the interpolated
    axes by default.

    This function transparently handles the entire path. An alternate approach would be to convert each segment
    separately and concatenate the interpolated axis with xarray.

    If the sentinel value 'G' for the Gamma point is included in the interpolation points, the coordinate axis of the
    interpolated coordinate will be shifted so that its value at the Gamma point is 0. You can opt out of this with the
    parameter 'shift_gamma'

    :param arr: Source data
    :param interpolation_points: Path vertices
    :param axis_name: Label for the interpolated axis. Under special circumstances a reasonable name will be chosen,
    such as when the interpolation dimensions are kx and ky: in this case the interpolated dimension will be labeled kp.
    In mixed or ambiguous situations the axis will be labeled by the default value 'inter'.
    :param resolution: Requested resolution along the interpolated axis.
    :param shift_gamma: Controls whether the interpolated axis is shifted to a value of 0 at Gamma.
    :param extend_to_edge: Controls whether or not to scale the vector S - G for symmetry point S so that you interpolate
    to the edge of the available data
    :param kwargs:
    :return: xr.DataArray containing the interpolated data.
    """

    if interpolation_points is None:
        raise ValueError('You must provide points specifying an interpolation path')

    def extract_symmetry_point(name):
        raw_point = arr.attrs['symmetry_points'][name]
        G = arr.attrs['symmetry_points']['G']

        if not extend_to_edge or name == 'G':
            return raw_point

        # scale the point so that it reaches the edge of the dataset
        S = np.array([raw_point[d] for d in arr.dims if d in raw_point])
        G = np.array([G[d] for d in arr.dims if d in raw_point])

        scale_factor = np.inf
        for i, d in enumerate([d for d in arr.dims if d in raw_point]):
            dS = (S - G)[i]
            coord = arr.coords[d]

            if np.abs(dS) < 0.001:
                continue

            if dS < 0:
                required_scale = (np.min(coord) - G[i]) / dS
                if required_scale < scale_factor:
                    scale_factor = float(required_scale)
            else:
                required_scale = (np.max(coord) - G[i]) / dS
                if required_scale < scale_factor:
                    scale_factor = float(required_scale)

        S = (S - G) * scale_factor + G
        return dict(zip([d for d in arr.dims if d in raw_point], S))


    parsed_interpolation_points = [
        x if isinstance(x, collections.Iterable) and not isinstance(x, str) else extract_symmetry_point(x)
        for x in interpolation_points
    ]

    free_coordinates = list(arr.dims)
    seen_coordinates = collections.defaultdict(set)
    for point in parsed_interpolation_points:
        for coord, value in point.items():
            seen_coordinates[coord].add(value)
            if coord in free_coordinates:
                free_coordinates.remove(coord)

    for point in parsed_interpolation_points:
        for coord, values in seen_coordinates.items():
            if coord not in point:
                if len(values) != 1:
                    raise ValueError('Ambiguous interpolation waypoint broadcast at dimension {}'.format(coord))
                else:
                    point[coord] = list(values)[0]

    if axis_name is None:
        axis_name = {
            ('beta', 'phi',): 'angle',
            ('chi', 'phi',): 'angle',
            ('phi', 'psi',): 'angle',
            ('phi', 'theta',): 'angle',
            ('kx', 'ky',): 'kp',
            ('kx', 'kz',): 'k',
            ('ky', 'kz',): 'k',
            ('kx', 'ky', 'kz',): 'k'
        }.get(tuple(sorted(seen_coordinates.keys())), 'inter')

        if axis_name == 'angle' or axis_name == 'inter':
            warnings.warn('Interpolating along axes with different dimensions '
                          'will not include Jacobian correction factor.')

    converted_coordinates = None
    converted_dims = free_coordinates + [axis_name]

    path_segments = list(zip(parsed_interpolation_points, parsed_interpolation_points[1:]))

    def element_distance(waypoint_a, waypoint_b):
        delta = np.array([waypoint_a[k] - waypoint_b[k] for k in waypoint_a.keys()])
        return np.linalg.norm(delta)

    def required_sampling_density(waypoint_a, waypoint_b):
        ks = waypoint_a.keys()
        dist = element_distance(waypoint_a, waypoint_b)
        delta = np.array([waypoint_a[k] - waypoint_b[k] for k in ks])
        delta_idx = [abs(d / (arr.coords[k][1] - arr.coords[k][0])) for d, k in zip(delta, ks)]
        return dist / np.max(delta_idx)

    # Approximate how many points we should use
    segment_lengths = [element_distance(*segment) for segment in path_segments]
    path_length = sum(segment_lengths)

    gamma_offset = 0 # offset the gamma point to a k coordinate of 0 if possible
    if 'G' in interpolation_points and shift_gamma:
        gamma_offset = sum(segment_lengths[0:interpolation_points.index('G')])

    if resolution is None:
        resolution = np.min([required_sampling_density(*segment) for segment in path_segments])

    def converter_for_coordinate_name(name):
        def raw_interpolator(*coordinates):
            return coordinates[free_coordinates.index(name)]

        if name in free_coordinates:
            return raw_interpolator

        # Conversion involves the interpolated coordinates
        def interpolated_coordinate_to_raw(*coordinates):
            # Coordinate order is [*free_coordinates, interpolated]
            interpolated = coordinates[len(free_coordinates)] + gamma_offset

            # Start with empty array that we will mask writes onto
            # We need to go with a masking approach rather than a concatenation based one because the coordinates
            # come from np.meshgrid
            dest_coordinate = np.zeros(shape=interpolated.shape)

            start = 0
            for i, l in enumerate(segment_lengths):
                end = start + l
                normalized = (interpolated - start) / l
                seg_start, seg_end = path_segments[i]
                dim_start, dim_end = seg_start[name], seg_end[name]
                mask = np.logical_and(normalized >= 0, normalized < 1)
                dest_coordinate[mask] = \
                    dim_start * (1 - normalized[mask]) + dim_end * normalized[mask]
                start = end

            return dest_coordinate

        return interpolated_coordinate_to_raw

    converted_coordinates = {d: arr.coords[d].values for d in free_coordinates}

    # Adjust this coordinate under special circumstances
    converted_coordinates[axis_name] = np.linspace(0, path_length, int(path_length / resolution)) - gamma_offset

    converted_ds = convert_coordinates(
        arr,
        converted_coordinates,
        {
            'dims': converted_dims,
            'transforms': dict(zip(arr.dims, [converter_for_coordinate_name(d) for d in arr.dims]))
        },
        as_dataset=True
    )

    if axis_name in arr.dims and len(parsed_interpolation_points) == 2:
        if parsed_interpolation_points[1][axis_name] < parsed_interpolation_points[0][axis_name]:
            # swap the sign on this axis as a convenience to the caller
            converted_ds.coords[axis_name].data = -converted_ds.coords[axis_name].data

    if 'id' in converted_ds.attrs:
        del converted_ds.attrs['id']
        provenance(converted_ds, arr, {
            'what': 'Slice along path',
            'by': 'slice_along_path',
            'parsed_interpolation_points': parsed_interpolation_points,
            'interpolation_points': interpolation_points,
        })

    return converted_ds


@update_provenance('Automatically k-space converted')
def convert_to_kspace(arr: xr.DataArray, forward=False, resolution=None, **kwargs):
    """
    "Forward" or "backward" converts the data to momentum space.

    "Backward"
    The standard method. Works in generality by regridding the data into the new coordinate space and then
    interpolating back into the original data.

    "Forward"
    By converting the coordinates, rather than by
    interpolating the data. As a result, the data will be totally unchanged by the conversion (if we do not
    apply a Jacobian correction), but the coordinates will no longer have equal spacing.

    This is only really useful for zero and one dimensional data because for two dimensional data, the coordinates
    must become two dimensional in order to fully specify every data point (this is true in generality, in 3D the
    coordinates must become 3D as well).

    The only exception to this is if the extra axes do not need to be k-space converted. As is the case where one
    of the dimensions is `cycle` or `delay`, for instance.

    :param arr:
    :return:
    """

    if isinstance(arr, xr.Dataset):
        warnings.warn('Remember to use a DataArray not a Dataset, attempting to extract spectrum')
        attrs = arr.attrs.copy()
        arr = normalize_to_spectrum(arr)
        arr.attrs.update(attrs)

    if forward:
        raise NotImplementedError('Forward conversion of datasets not supported. Coordinate conversion is. '
                                  'See `arpes.utilities.conversion.forward.convert_coordinates_to_kspace_forward`')

    has_eV = 'eV' in arr.dims

    # TODO be smarter about the resolution inference
    old_dims = list(deepcopy(arr.dims))
    remove_dims = ['eV', 'delay', 'cycle', 'temp', 'x', 'y']

    def unconvertible(dimension: str) -> bool:
        if dimension in remove_dims:
            return True

        if 'volt' in dimension:
            return True

        return False

    removed = []

    for to_remove in arr.dims:
        if unconvertible(to_remove):
            removed.append(to_remove)
            old_dims.remove(to_remove)

    # This should always be true because otherwise we have no hope of carrying
    # through with the conversion
    if 'eV' in removed:
        removed.remove('eV') # This is put at the front as a standardization

    old_dims.sort()

    if not old_dims:
        return arr # no need to convert, might be XPS or similar

    converted_dims = (['eV'] if has_eV else []) + {
        ('phi',): ['kp'],

        ('phi', 'theta'): ['kx', 'ky'],
        ('beta', 'phi'): ['kx', 'ky'],
        ('phi', 'psi'): ['kx', 'ky'],

        ('hv', 'phi'): ['kp', 'kz'],

        ('hv', 'phi', 'theta'): ['kx', 'ky', 'kz'],
        ('beta', 'hv', 'phi'): ['kx', 'ky', 'kz'],
        ('hv', 'phi', 'psi'): ['kx', 'ky', 'kz'],
    }.get(tuple(old_dims)) + removed

    convert_cls = {
        ('phi',): ConvertKp,

        ('beta', 'phi'): ConvertKxKy,
        ('phi', 'theta'): ConvertKxKy,
        ('phi', 'psi'): ConvertKxKy,
        #('chi', 'phi',): ConvertKxKy,

        ('hv', 'phi'): ConvertKpKz,
    }.get(tuple(old_dims))
    converter = convert_cls(arr, converted_dims)

    n_kspace_coordinates = len(set(converted_dims).intersection({'kp', 'kx', 'ky', 'kz'}))
    if n_kspace_coordinates > 1 and forward:
        raise AnalysisError('You cannot forward convert more than one momentum to k-space.')

    converted_coordinates = converter.get_coordinates(resolution)

    return convert_coordinates(
        arr, converted_coordinates, {
            'dims': converted_dims,
            'transforms': dict(zip(arr.dims, [converter.conversion_for(d) for d in arr.dims]))})[0]


def convert_coordinates(arr: xr.DataArray, target_coordinates, coordinate_transform, as_dataset=False):
    ordered_source_dimensions = arr.dims
    grid_interpolator = grid_interpolator_from_dataarray(
        arr.transpose(*ordered_source_dimensions), fill_value=float('nan'))

    # Skip the Jacobian correction for now
    # Convert the raw coordinate axes to a set of gridded points
    meshed_coordinates = np.meshgrid(*[target_coordinates[dim] for dim in coordinate_transform['dims']],
                                     indexing='ij')
    meshed_coordinates = [meshed_coord.ravel() for meshed_coord in meshed_coordinates]

    if 'eV' not in arr.dims:
        meshed_coordinates = [arr.S.lookup_offset_coord('eV')] + meshed_coordinates

    old_coord_names = [dim for dim in arr.dims if dim not in target_coordinates]
    old_coordinate_transforms = [coordinate_transform['transforms'][dim] for dim in arr.dims if dim not in target_coordinates]
    old_dimensions = [np.reshape(tr(*meshed_coordinates), [len(target_coordinates[d]) for d in coordinate_transform['dims']], order='C')
                      for tr in old_coordinate_transforms]

    ordered_transformations = [coordinate_transform['transforms'][dim] for dim in arr.dims]
    converted_volume = grid_interpolator(np.array([tr(*meshed_coordinates) for tr in ordered_transformations]).T)

    # Wrap it all up
    def acceptable_coordinate(c: Union[np.ndarray, xr.DataArray]) -> bool:
        # Currently we do this to filter out coordinates that are functions of the old angular dimensions,
        # we could forward convert these, but right now we do not
        try:
            if set(c.dims).issubset(coordinate_transform['dims']):
                return True
            else:
                return False
        except:
            return True

    target_coordinates = {k: v for k, v in target_coordinates.items() if acceptable_coordinate(v)}
    data = xr.DataArray(
        np.reshape(converted_volume, [len(target_coordinates[d]) for d in coordinate_transform['dims']], order='C'),
        target_coordinates,
        coordinate_transform['dims'],
        attrs=arr.attrs,
    )
    old_mapped_coords = [xr.DataArray(values, target_coordinates, coordinate_transform['dims'], attrs=arr.attrs) for values in
                         old_dimensions]
    if as_dataset:
        vars = {'data': data}
        vars.update(dict(zip(old_coord_names, old_mapped_coords)))
        return xr.Dataset(vars, attrs=arr.attrs)

    return data, old_mapped_coords



import math

import numpy as np

import xarray as xr
from arpes.provenance import update_provenance

from .axis_preparation import transform_dataarray_axis

__all__ = ['build_KE_coords_to_time_pixel_coords', 'build_KE_coords_to_time_coords', 'process_DLD', 'process_SToF',]


@update_provenance('Convert ToF data from timing signal to kinetic energy')
def convert_to_kinetic_energy(dataarray, kinetic_energy_axis):
    """
    Convert the ToF timing information into an energy histogram

    The core of these routines come from the Igor procedures in
    ``LoadTOF7_3.51.ipf``.

    To explain in a bit more detail what actaully happens, this
    function essentially

    1. Determines the time to energy conversion factor
    2. Calculates the requested energy binning range and allocates arrays
    3. Rebins a time spectrum into an energy spectrum, preserving the
       spectral weight, this requires a modicum of care around splitting
       counts at the edges of the new bins.
    """

    # This should be simplified
    # c = (0.5) * (9.11e-31) * self.mstar * (self.length ** 2) / (1.6e-19) * (1e18)
    # Removed factors of ten and substituted mstar = 0.5
    c = (0.5) * (9.11e6) * dataarray.S.spectrometer['mstar'] * (dataarray.S.spectrometer['length'] ** 2) / 1.6

    new_dim_order = list(dataarray.dims)
    new_dim_order.remove('time')
    new_dim_order = ['time'] + new_dim_order
    dataarray = dataarray.transpose(*new_dim_order)
    new_dim_order[0] = 'eV'

    timing = dataarray.coords['time'].values
    assert timing[1] > timing[0]
    t_min, t_max = np.min(timing), np.max(timing)

    # Prep arrays
    d_energy = kinetic_energy_axis[1] - kinetic_energy_axis[0]
    time_index = 0 # after transpose
    new_shape = list(dataarray.data.shape)
    new_shape[time_index] = len(kinetic_energy_axis)

    new_data = np.zeros(tuple(new_shape))

    def energy_to_time(conv, energy):
        return math.sqrt(conv / energy)

    # Rebin data
    old_data = dataarray.data
    for i, E in enumerate(kinetic_energy_axis):
        t_L = energy_to_time(c, E + d_energy / 2)
        t_S = energy_to_time(c, E - d_energy / 2)

        # clamp
        t_L = t_L if t_L <= t_max else t_max
        t_S = t_S if t_S > t_min else t_min

        # with some math we could back calculate the index, but we don't need to
        t_L_idx = np.searchsorted(timing, t_L)
        t_S_idx = np.searchsorted(timing, t_S)

        new_data[i] = np.sum(old_data[t_L_idx:t_S_idx], axis=0) + \
                      ((timing[t_L_idx] - t_L) * old_data[t_L_idx]) + \
                      ((t_S - timing[t_S_idx - 1]) * old_data[t_S_idx - 1]) / d_energy

    new_coords = dict(dataarray.coords)
    del new_coords['time']
    new_coords['eV'] = kinetic_energy_axis

    # Put provenance here

    return xr.DataArray(
        new_data,
        coords=new_coords,
        dims=new_dim_order,
        attrs=dataarray.attrs,
        name=dataarray.name
    )


def build_KE_coords_to_time_pixel_coords(dataset: xr.Dataset, interpolation_axis):
    conv = dataset.S.spectrometer['mstar'] * (9.11e6) * 0.5 * (dataset.S.spectrometer['length'] ** 2) / 1.6
    time_res = 0.17 # this is only approximate

    def KE_coords_to_time_pixel_coords(coords, axis=None):
        """
        Like ``KE_coords_to_time_coords`` but it converts to the raw timing pixels off of
        a DLD instead to the unitful values that we receive from the Spin-ToF DAQ
        :param coords: tuple of coordinates
        :return: new tuple of converted coordinates
        """
        kinetic_energy_pixel = coords[axis]
        kinetic_energy = interpolation_axis[kinetic_energy_pixel]
        real_timing = math.sqrt(conv / kinetic_energy)
        pixel_timing = (real_timing - dataset.attrs['timing_offset']) / time_res
        coords_list = list(coords)
        coords_list[axis] = pixel_timing

        return tuple(coords_list)
    return KE_coords_to_time_pixel_coords


def build_KE_coords_to_time_coords(dataset: xr.Dataset, interpolation_axis):
    """
    Geometric transform assumes pixel -> pixel transformations, so we need to get the index associated
    to the appropriate timing value
    :param dataset:
    :param interpolation_axis:
    :return:
    """
    conv = dataset.S.spectrometer['mstar'] * (9.11e6) * 0.5 * (dataset.S.spectrometer['length'] ** 2) / 1.6
    timing = dataset.coords['time']
    photon_offset = dataset.attrs['laser_t0'] + dataset.S.spectrometer['length'] * (10 / 3)
    low_offset = np.min(timing)
    d_timing = timing[1] - timing[0]

    def KE_coords_to_time_coords(coords, axis=None):
        """
        Used to convert the timing coordinates off of the spin-ToF to kinetic energy coordinates.
        As the name suggests, because of how scipy.ndimage interpolates, we require the inverse
        coordinate transform.

        All coordinates except for the energy coordinate are left untouched.

        We do the same logic done implicitly in timeProcessX in order to get the part of the data that has time
        coordinate less than the nominal t0. This is necessary because the recorded times are the time between electron
        events and laser pulses, rather than the other way around.

        :param coords: tuple of coordinates
        :return: new tuple of converted coordinates
        """
        kinetic_energy = interpolation_axis[coords[axis]]
        real_timing = math.sqrt(conv / kinetic_energy)
        real_timing = photon_offset - real_timing
        coords_list = list(coords)
        coords_list[axis] = len(timing) - (real_timing - low_offset) / d_timing
        return tuple(coords_list)

    return KE_coords_to_time_coords


@update_provenance('Convert ToF data from timing signal to kinetic energy ALT')
def convert_SToF_to_energy(dataset: xr.Dataset):
    """
    Achieves the same computation as timeProcessX and t2energyProcessX in LoadTOF_3.51.ipf
    :param dataset:
    :return:
    """
    e_min, e_max = 0.1, 10.

    # TODO, we can better infer a reasonable gridding here
    spacing = dataset.attrs.get('dE', 0.005)
    ke_axis = np.linspace(e_min, e_max, int((e_max - e_min) / spacing))

    drs = {k: v for k, v in dataset.data_vars.items() if 'time' in v.dims}

    new_dataarrays = [convert_to_kinetic_energy(dr, ke_axis) for dr in drs.values()]

    for v in new_dataarrays:
        dataset[v.name.replace('t_', '')] = v

    return dataset


@update_provenance('Preliminary data processing Spin-ToF')
def process_SToF(dataset: xr.Dataset):
    """
    This isn't the best unit conversion function because it doesn't properly
    take into account the Jacobian of the coordinate conversion. This can
    be fixed by multiplying each channel by the appropriate ammount, but it might still
    be best to use the alternative method.

    :param dataset:
    :return:
    """
    e_min = dataset.attrs.get('E_min', 1)
    e_max = dataset.attrs.get('E_max', 10)
    de = dataset.attrs.get('dE', 0.01)
    ke_axis = np.linspace(e_min, e_max, (e_max - e_min) / de)

    dataset = transform_dataarray_axis(
        build_KE_coords_to_time_coords(dataset, ke_axis),
        'time', 'eV', ke_axis, dataset, lambda x: x,
    )

    dataset = dataset.rename({'t_up': 'up', 't_down': 'down'})

    if 'up' in dataset.data_vars:
        # apply the sherman function corrections
        sherman = dataset.attrs.get('sherman', 0.2)
        polarization = 1/sherman * (dataset.up - dataset.down)/(dataset.up + dataset.down)
        new_up = (dataset.up + dataset.down) * (1 + polarization)
        new_down = (dataset.up + dataset.down)* (1 - polarization)
        dataset = dataset.assign(up=new_up, down=new_down)

    return dataset


@update_provenance('Preliminary data processing prototype DLD')
def process_DLD(dataset: xr.Dataset):
    e_min = 1
    ke_axis = np.linspace(e_min, dataset.attrs['E_max'], (dataset.attrs['E_max'] - e_min) / dataset.attrs['dE'])
    dataset = transform_dataarray_axis(
        build_KE_coords_to_time_pixel_coords(dataset, ke_axis),
        't_pixels', 'kinetic', ke_axis, dataset, lambda x: 'kinetic_spectrum'
    )

    return dataset



import itertools
import os.path
import subprocess

__all__ = ('shim_wave_note',)


def shim_wave_note(path):
    """
    Hack to read the corrupted wavenote out of the h5 files that Igor has been producing.
    h5 dump still produces the right value, so we use it from the command line in order to get the value of the note.

    This is not necessary unless you are trying to read HDF files exported from Igor (the
    preferred way before we developed an Igor data loading plugin).

    :param path: Location of the file
    :return:
    """
    wave_name = os.path.splitext(os.path.basename(path))[0]
    cmd = 'h5dump -A --attribute /{}/IGORWaveNote {}'.format(wave_name, path)
    h5_out = subprocess.getoutput(cmd)

    split_data = h5_out[h5_out.index('DATA {'):]
    assert len(split_data.split('"')) == 3
    data = split_data.split('"')[1]

    # remove stuff below the end of the header
    try:
        data = data[:data.index('ENDHEADER')]
    except ValueError:
        pass

    lines = [l.strip() for l in data.splitlines() if '=' in l]
    lines = itertools.chain(*[l.split(',') for l in lines])
    return dict([l.split('=') for l in lines])



"""
Some bespoke exceptions that can be used in control sequences and to
provide more information to the user. I (Conrad) prefer to use warnings for
the latter purpose, but there are reasons to throw these errors in a variety of circumstances.
"""

class AnalysisError(Exception):
    """
    Base class to indicate that something scientific went wrong and
    was not handled in an appropriate way.

    # Examples

    1. A bad fit from scipy.optimize in an internal function or analysis
    routine that could not be handled by the user
    """


class AnalysisWarning(UserWarning):
    """
    Non-fatal, but the user probably forgot something really significant that
    invalidates any science that follows
    """


class DataPreparationError(AnalysisError):
    """
    Indicates that the user needs to perform some data preparation step
    before the analysis can proceed.
    """


class ConfigurationError(Exception):
    """
    Indicates that the user needs to supply more configuration. This could be
    setting some directories in which to place plots, or indicating what the current
    workspace being used (i.e. where to look for datasets and data) is.
    """



# pylint: disable=import-error

from PyQt5 import QtGui, QtWidgets

__all__ = ('AxisInfoWidget',)

class AxisInfoWidget(QtWidgets.QGroupBox):
    def __init__(self, parent=None, root=None, axis_index=None):
        super().__init__(title=str(axis_index), parent=parent)

        self.layout = QtGui.QGridLayout(self)

        self.label = QtWidgets.QLabel('Cursor: ')
        self.transpose_button = QtWidgets.QPushButton('To Front')
        self.transpose_button.clicked.connect(self.on_transpose)

        self.layout.addWidget(self.label)
        self.layout.addWidget(self.transpose_button)

        self.axis_index = axis_index
        self.root = root
        self.setLayout(self.layout)
        self.recompute()

    def recompute(self):
        self.setTitle(self.root.data.dims[self.axis_index])
        try:
            cursor_index = self.root.context['cursor'][self.axis_index]
            cursor_value = self.root.context['value_cursor'][self.axis_index]
            self.label.setText('Cursor: {}, {:.3g}'.format(int(cursor_index), cursor_value))
        except KeyError:
            pass

    def on_transpose(self):
        try:
            self.root.transpose_to_front(self.axis_index)
        except Exception:
            pass



"""
Useful constants for experiments and some useful science facts
Much of this is collected from past students, especially Jeff's 'Cstes.ipf'

Some of this will disappear in future updates, as we move away from magic constants towards
bundling necessary information on endstation classes.
"""

from numpy import pi

MODE_ARPES = 'arpes'
MODE_TRARPES = 'trarpes'
MODE_SARPES = 'sarpes'
MODE_STARPES = 'starpes'

EXPERIMENT_MODES = [
    MODE_ARPES,
    MODE_TRARPES,
    MODE_SARPES,
    MODE_STARPES,
]

TIME_RESOLVED_MODES = [
    MODE_TRARPES,
    MODE_STARPES,
]

SPIN_RESOLVED_MODES = [
    MODE_SARPES,
    MODE_STARPES,
]

def mode_has_spin_resolution(mode):
    return mode in SPIN_RESOLVED_MODES

def mode_has_time_resolution(mode):
    return mode in TIME_RESOLVED_MODES

LATTICE_CONSTANTS = {
    'Bi-2212': 3.83,
    'NCCO': 3.942,
    'Hg-2201': 3.8797,
    'BaFe2As2': 3.9625,
}

# eV, A reasonablish value if you aren't sure for the particular sample
WORK_FUNCTION = 4.38

HBAR = 1.05 * 10**(-34)
HBAR_EV = 6.52 * 10**(-16)
BARE_ELECTRON_MASS = 9.109e-31 # kg
HBAR_SQ_EV_PER_ELECTRON_MASS = 0.475600805657 # hbar^2 / m0 in eV^2 s^2 / kg
HBAR_SQ_EV_PER_ELECTRON_MASS_ANGSTROM_SQ = 7.619964 # (hbar^2) / (m0 * angstrom ^2) in eV

K_BOLTZMANN_EV_KELVIN = 8.61733e-5 # in units of eV / Kelvin
K_BOLTZMANN_MEV_KELVIN = 1000 * K_BOLTZMANN_EV_KELVIN # meV / Kelvin

HC = 1239.84172 # in units of eV * nm

HEX_ALPHABET = "ABCDEF0123456789"
ALPHABET = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
ALPHANUMERIC = "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"

STRAIGHT_TOF_LENGTH = 0.937206
SPIN_TOF_LENGTH = 1.1456
DLD_LENGTH = 1.1456 # This isn't correct but it should be a reasonable guess

K_INV_ANGSTROM = 0.5123

SPECTROMETER_MC = {
    'name': 'MC',
    'rad_per_pixel': (1/10) * (pi / 180),
    'type': 'hemisphere',
    'is_slit_vertical': False,
}

SPECTROMETER_MC_OLD = {
    'name': 'MC_OLD',
    'type': 'hemisphere',
    'rad_per_pixel': 0.125 * (pi / 180),
    'is_slit_vertical': False,
}

SPECTROMETER_STRAIGHT_TOF = {
    'name': 'STRAIGHT_ToF',
    'length': STRAIGHT_TOF_LENGTH,
    'mstar': 1.0,
    'type': 'tof',
    'dof': ['t'],
    'scan_dof': ['theta'],
}

SPECTROMETER_SPIN_TOF = {
    'name': 'SPIN_ToF',
    'length': SPIN_TOF_LENGTH,
    'mstar': 0.5,
    'type': 'tof',
    'dof': ['time', 'spin'],
    'scan_dof': ['theta', 'beta'],
}

SPECTROMETER_DLD = {
    'name': 'DLD',
    'length': DLD_LENGTH,
    'type': 'tof',
    'dof_type': {
        'timing': ['x_pixels', 't_pixels'],
        'spatial': ['x_pixels', 'y_pixels'],
    },
    'scan_dof': ['theta'],
}

SPECTROMETER_BL4 = {
    'name': 'BL4',
    'is_slit_vertical': True,
    'type': 'hemisphere',
    'dof': ['theta', 'sample_phi'],
}

SPECTROMETER_BL7 = {
    'name': 'BL7',
    'is_slit_vertical': True,
    'type': 'hemisphere',
    'dof': ['theta', 'sample_phi'],
}

SPECTROMETER_ANTARES = {
    'name': 'ANTARES',
    'is_slit_vertical': True,
    'type': 'hemisphere',
    'dof': ['theta', 'sample_phi'],
}

SPECTROMETER_KAINDL = {
    'name': 'Kaindl',
    'is_slit_vertical': True,
    'type': 'hemisphere',
    'dof': ['theta', 'sample_phi'],
}

HV_CONVERSION = 3.81

FINE_K_GRAINING = 0.01
MEDIUM_FINE_K_GRAINING = 0.02
MEDIUM_K_GRAINING = 0.05
COARSE_K_GRAINING = 0.1



import matplotlib
import numpy as np
from matplotlib import pyplot as plt
from matplotlib import colors, gridspec

from arpes.analysis.xps import approximate_core_levels
from arpes.provenance import save_plot_provenance
from arpes.utilities import normalize_to_spectrum

from .utils import path_for_plot, savefig

__all__ = ('plot_dos', 'plot_core_levels',)


@save_plot_provenance
def plot_core_levels(data, title=None, out=None, norm=None, dos_pow=1, core_levels=None, binning=1, promenance=5, **kwargs):
    axes, cbar = plot_dos(data=data, title=title, out=None, norm=norm, dos_pow=dos_pow, **kwargs)

    if core_levels is None:
        core_levels = approximate_core_levels(data, binning=binning, promenance=promenance)

    for core_level in core_levels:
        axes[1].axvline(core_level, color='red', ls='--')

    if out is not None:
        savefig(out, dpi=400)
        return path_for_plot(out)
    else:
        return axes, cbar


@save_plot_provenance
def plot_dos(data, title=None, out=None, norm=None, dos_pow=1, **kwargs):
    data = normalize_to_spectrum(data)

    fig = plt.figure(figsize=(14, 6))
    fig.subplots_adjust(hspace=0.00)
    gs = gridspec.GridSpec(2, 1, height_ratios=[3, 1])

    ax0 = plt.subplot(gs[0])
    axes = (ax0, plt.subplot(gs[1], sharex=ax0))

    data.values[np.isnan(data.values)] = 0
    cbar_axes = matplotlib.colorbar.make_axes(axes, pad=0.01)

    mesh = data.plot(ax=axes[0], norm=norm or colors.PowerNorm(gamma=0.15))

    axes[1].set_facecolor((0.95, 0.95, 0.95))
    density_of_states = data.S.sum_other(['eV'])
    (density_of_states ** dos_pow).plot(ax=axes[1])

    cbar = plt.colorbar(mesh, cax=cbar_axes[0])
    cbar.set_label('Photoemission Intensity (Arb.)')

    axes[1].set_ylabel('Spectrum DOS', labelpad=12)
    axes[0].set_title(title or '')

    if out is not None:
        savefig(out, dpi=400)
        return path_for_plot(out)
    else:
        return fig, axes, cbar



import re
import os
from pathlib import Path

import numpy as np
import pandas as pd

import typing
import xarray as xr

import arpes
from arpes.endstations import HemisphericalEndstation, SESEndstation

__all__ = ('KaindlEndstation',)


def find_kaindl_files_associated(reference_path: Path):
    name_match = re.match(r'([\w+]+_scan_[0-9][0-9][0-9]_)[0-9][0-9][0-9]\.pxt', reference_path.name)

    if name_match is None:
        return [reference_path]

    # otherwise need to collect all of the components
    fragment = name_match.groups()[0]
    components = list(reference_path.parent.glob('{}*.pxt'.format(fragment)))
    components.sort()

    return components


class KaindlEndstation(HemisphericalEndstation, SESEndstation):
    """
    The Robert Kaindl Endstation For HHG ARPES
    """

    PRINCIPAL_NAME = 'Kaindl'
    ALIASES = []

    RENAME_KEYS = {
        'Delay Stage': 'delay',
    }

    def resolve_frame_locations(self, scan_desc: dict = None):
        if scan_desc is None:
            raise ValueError('Must pass dictionary as file scan_desc to all endstation loading code.')

        original_data_loc = scan_desc.get('path', scan_desc.get('file'))
        p = Path(original_data_loc)
        if not p.exists():
            original_data_loc = os.path.join(arpes.config.DATA_PATH, original_data_loc)

        p = Path(original_data_loc)
        return find_kaindl_files_associated(p)

    def concatenate_frames(self, frames=typing.List[xr.Dataset], scan_desc: dict = None):
        if len(frames) < 2:
            return super().concatenate_frames(frames)

        # determine which axis to stitch them together along, and then do this
        original_filename = scan_desc.get('path', scan_desc.get('file'))

        internal_match = re.match(r'([a-zA-Z0-9\w+_]+)_[0-9][0-9][0-9]\.pxt', Path(original_filename).name)
        if internal_match.groups():
            motors_path = str(Path(original_filename).parent / '{}_Motor_Pos.txt'.format(internal_match.groups()[0]))
            try:
                with open(motors_path, 'r') as f:
                    lines = f.readlines()

                axis_name = lines[0].strip()
                axis_name = self.RENAME_KEYS.get(axis_name, axis_name)
                values = [float(l.strip()) for l in lines[1:len(frames) + 1]]

                for v, f in zip(values, frames):
                    f.coords[axis_name] = v

                frames.sort(key=lambda x: x.coords[axis_name])
                return xr.concat(frames, axis_name)
            except Exception:
                pass

    def postprocess_final(self, data: xr.Dataset, scan_desc: dict = None):
        original_filename = scan_desc.get('path', scan_desc.get('file'))
        internal_match = re.match(r'([a-zA-Z0-9\w+_]+_[0-9][0-9][0-9])\.pxt', Path(original_filename).name)
        all_filenames = find_kaindl_files_associated(Path(original_filename))
        all_filenames = ['{}\\{}_AI.txt'.format(f.parent, f.stem) for f in all_filenames]

        def load_attr_for_frame(filename, attr_name):
            df = pd.read_csv(filename, sep='\t', skiprows=6)
            return np.mean(df[attr_name])

        def attach_attr(data, attr_name, as_name):
            photocurrents = np.array([load_attr_for_frame(f, attr_name)
                                      for f in all_filenames])

            if len(photocurrents) == 1:
                data[as_name] = photocurrents[0]
            else:

                non_spectrometer_dims = [d for d in data.spectrum.dims if d not in {'eV', 'phi'}]
                non_spectrometer_coords = {c: v for c, v in data.spectrum.coords.items()
                                           if c in non_spectrometer_dims}

                new_shape = [len(data.coords[d]) for d in non_spectrometer_dims]
                photocurrent_arr = xr.DataArray(
                    photocurrents.reshape(new_shape), coords=non_spectrometer_coords, dims=non_spectrometer_dims)

                data = xr.merge([data, xr.Dataset(dict([[as_name, photocurrent_arr]]))])

            return data

        try:
            data = attach_attr(data, 'Photocurrent', 'photocurrent')
            data = attach_attr(data, 'Temperature B', 'temp')
            data = attach_attr(data, 'Temperature A', 'cryotip_temp')
        except FileNotFoundError as e:
            print(e)

        if internal_match.groups():
            attrs_path = str(Path(original_filename).parent / '{}_AI.txt'.format(internal_match.groups()[0]))

            try:
                extra = pd.read_csv(attrs_path, sep='\t', skiprows=6)
                data = data.assign_attrs(extra=extra.to_json())
            except Exception:
                # WELP we tried
                pass

        deg_to_rad_coords = {'theta', 'beta', 'phi'}

        for c in deg_to_rad_coords:
            if c in data.dims:
                data.coords[c] = data.coords[c] * np.pi / 180

        deg_to_rad_attrs = {'theta', 'beta', 'alpha', 'chi'}
        for angle_attr in deg_to_rad_attrs:
            if angle_attr in data.attrs:
                data.attrs[angle_attr] = float(data.attrs[angle_attr]) * np.pi / 180

        data = super().postprocess_final(data, scan_desc)

        return data



import re
from collections import OrderedDict

from arpes.utilities.xarray import lift_dataarray_attrs, lift_datavar_attrs
from typing import Any, Dict, Union

__all__ = ('rename_keys', 'clean_keys', 'rename_dataarray_attrs', 'rename_datavar_attrs',
           'clean_datavar_attribute_names', 'clean_attribute_names', 'case_insensitive_get')

def _rename_key(d: Union[Dict[str, Any], OrderedDict], k: str, nk: str) -> None:
    if k in d:
        d[nk] = d[k]
        del d[k]


def rename_keys(d: Union[Dict[str, Any], OrderedDict], keys_dict: Dict[str, str]) -> Union[Dict[str, Any], OrderedDict]:
    d = d.copy()
    for k, nk in keys_dict.items():
        _rename_key(d, k, nk)

    return d


def clean_keys(d):
    def clean_single_key(k):
        k = k.replace(' ', '_')
        k = k.replace('.', '_')
        k = k.lower()
        k = re.sub(r'[()/?]', '', k)
        k = k.replace('__', '_')
        return k

    return dict(zip([clean_single_key(k) for k in d.keys()], d.values()))


def case_insensitive_get(d: dict, key: str, default=None, take_first=False):
    """
    Looks up a key in a dictionary ignoring case. We use this sometimes to be
    nicer to users who don't provide perfectly sanitized data
    :param d:
    :param key:
    :param default:
    :param take_first:
    :return:
    """
    found_value = False
    value = None

    for k, v in d.items():
        if k.lower() == key.lower():
            if not take_first and found_value:
                raise ValueError('Duplicate case insensitive keys')

            value = v
            found_value = True

            if take_first:
                break

    if not found_value:
        return default

    return value


rename_dataarray_attrs = lift_dataarray_attrs(rename_keys)
clean_attribute_names = lift_dataarray_attrs(clean_keys)

rename_datavar_attrs = lift_datavar_attrs(rename_keys)
clean_datavar_attribute_names = lift_datavar_attrs(clean_keys)



import copy

import numpy as np
from scipy import ndimage

import xarray as xr
from arpes.provenance import provenance

__all__ = ('gaussian_filter_arr', 'gaussian_filter', 'boxcar_filter_arr', 'boxcar_filter',)


def gaussian_filter_arr(arr: xr.DataArray, sigma=None, n=1, default_size=1):
    """
    Functionally wraps scipy.ndimage.filters.gaussian_filter with the advantage that the sigma
    is coordinate aware.

    :param arr:
    :param sigma: Kernel sigma, specified in terms of axis units. An axis that is not specified
                  will have a kernel width of `default_size` in index units.
    :param n: Repeats n times.
    :param default_size: Changes the default kernel width for axes not specified in `sigma`. Changing this
                         parameter and leaving `sigma` as None allows you to smooth with an even-width
                         kernel in index-coordinates.
    :return: xr.DataArray: smoothed data.
    """
    if sigma is None:
        sigma = {}

    sigma = {k: int(v / (arr.coords[k][1] - arr.coords[k][0])) for k, v in sigma.items()}
    for dim in arr.dims:
        if dim not in sigma:
            sigma[dim] = default_size

    sigma = tuple(sigma[k] for k in arr.dims)

    values = arr.values
    for _ in range(n):
        values = ndimage.filters.gaussian_filter(values, sigma)

    filtered_arr = xr.DataArray(
        values,
        arr.coords,
        arr.dims,
        attrs=copy.deepcopy(arr.attrs)
    )

    if 'id' in filtered_arr.attrs:
        del filtered_arr.attrs['id']

        provenance(filtered_arr, arr, {
            'what': 'Gaussian filtered data',
            'by': 'gaussian_filter_arr',
            'sigma': sigma,
        })

    return filtered_arr


def gaussian_filter(sigma=None, n=1):
    """
    A partial application of `gaussian_filter_arr` that can be passed to derivative analysis functions.
    :param sigma:
    :param n:
    :return:
    """

    def f(arr):
        return gaussian_filter_arr(arr, sigma, n)

    return f


def boxcar_filter(size=None, n=1):
    """
    A partial application of `boxcar_filter_arr` that can be passed to derivative analysis functions.
    :param size:
    :param n:
    :return:
    """

    def f(arr):
        return boxcar_filter_arr(arr, size, n)

    return f


def boxcar_filter_arr(arr: xr.DataArray, size=None, n=1, default_size=1, skip_nan=True):
    """
    Functionally wraps scipy.ndimage.filters.gaussian_filter with the advantage that the sigma
    is coordinate aware.

    :param arr:
    :param size: Kernel size, specified in terms of axis units. An axis that is not specified
                 will have a kernel width of `default_size` in index units.
    :param n: Repeats n times.
    :param default_size: Changes the default kernel width for axes not specified in `sigma`. Changing this
                         parameter and leaving `sigma` as None allows you to smooth with an even-width
                         kernel in index-coordinates.
    :param skip_nan: By default, masks parts of the data which are NaN to prevent poor filter results.
    :return: xr.DataArray: smoothed data.
    """

    if size is None:
        size = {}

    size = {k: int(v / (arr.coords[k][1] - arr.coords[k][0])) for k, v in size.items()}
    for dim in arr.dims:
        if dim not in size:
            size[dim] = default_size

    size = tuple(size[k] for k in arr.dims)

    if skip_nan:
        nan_mask = np.copy(arr.values) * 0 + 1
        nan_mask[arr.values != arr.values] = 0
        filtered_mask = ndimage.filters.uniform_filter(nan_mask, size)

        values = np.copy(arr.values)
        values[values != values] = 0

        for _ in range(n):
            values = ndimage.filters.uniform_filter(values, size) / filtered_mask
            values[nan_mask == 0] = 0
    else:
        for i in range(n):
            values = ndimage.filters.uniform_filter(values, size)

    filtered_arr = xr.DataArray(
        values,
        arr.coords,
        arr.dims,
        attrs=copy.deepcopy(arr.attrs)
    )

    if 'id' in arr.attrs:
        del filtered_arr.attrs['id']

        provenance(filtered_arr, arr, {
            'what': 'Boxcar filtered data',
            'by': 'boxcar_filter_arr',
            'size': size,
            'skip_nan': skip_nan,
        })

    return filtered_arr



from arpes.analysis import rebin
from arpes.fits import GStepBModel, broadcast_model
from arpes.provenance import update_provenance
from arpes.typing import DataType
from arpes.utilities.normalize import normalize_to_spectrum

__all__ = ('fs_gap',)


@update_provenance('Broadcast Fermi Surface Gap')
def fs_gap(data: DataType, shape=None, energy_range=None):
    data = normalize_to_spectrum(data)

    if energy_range is None:
        energy_range = slice(-0.1, None)

    data.sel(eV=energy_range)

    reduction = None
    if shape is None:
        # Just rebin the data along 'phi'
        reduction = {'phi': 16}

    data = rebin(data, reduction=reduction, shape=shape)
    return broadcast_model(GStepBModel, data, ['phi', 'beta'])



"""
Specialized type annotations for use in PyARPES.

In particular, we frequently allow using the `DataType` annotation,
which refers to either an xarray.DataArray or xarray.Dataset.

Additionally, we often use `NormalizableDataType` which
means essentially anything that can be turned into a dataset,
for instance by loading from the cache using an ID, or which is
literally already data.
"""

import uuid

import typing
import xarray as xr

__all__ = ['DataType', 'NormalizableDataType', 'xr_types']

DataType = typing.Union[xr.DataArray, xr.Dataset]
NormalizableDataType = typing.Union[DataType, str, uuid.UUID]

xr_types = (xr.DataArray, xr.Dataset,)



# pylint: disable=redefined-outer-name

import os
import warnings

from arpes.endstations import load_scan
from arpes.io import dataset_exists, save_dataset, save_dataset_for_export
from arpes.utilities import (attach_extra_dataset_columns,
                             clean_datavar_attribute_names,
                             modern_clean_xlsx_dataset,
                             rename_datavar_standard_attrs)
from arpes.utilities.dataset import walk_datasets

__all__ = ('prepare_raw_files',)


def prepare_raw_files(workspace=None, debug=False, reload=False, file=None, quiet=False, export=False, **kwargs):
    import arpes.xarray_extensions
    import arpes.config

    stop_debug = False
    if debug == 'first':
        stop_debug = True
        debug = True

    arpes.config.attempt_determine_workspace(workspace)
    assert 'path' in arpes.config.CONFIG['WORKSPACE']
    workspace_path = arpes.config.CONFIG['WORKSPACE']['path']

    print('Found: {}'.format(workspace_path))
    files = walk_datasets(use_workspace=True)
    if file:
        print("├{}".format(file))
        files = [os.path.join(workspace_path, file)]

    for dataset_path in files:
        # we used to pass "use_soft_match" here, we are switching to a regex based approach so this might
        # initially cause some problems but should be much better in the end
        ds = modern_clean_xlsx_dataset(dataset_path, with_inferred_cols=False, write=True)

        print('└┐')
        for iter_index, (file, scan) in enumerate(ds.iterrows()):
            print(' ├{}'.format(file))
            scan['file'] = scan.get('path', file)
            if not dataset_exists(scan.get('id')) or reload:
                try:
                    with warnings.catch_warnings():
                        if quiet:
                            warnings.simplefilter('ignore')
                        data = load_scan(dict(scan), **kwargs)
                    data = rename_datavar_standard_attrs(data)
                    data = clean_datavar_attribute_names(data)
                    if export:
                        index = scan['file']
                        try:
                            index = int(index)
                        except ValueError:
                            pass
                        if not isinstance(index, int):
                            index = iter_index + 1
                        save_dataset_for_export(data, index, force=True)
                    else:
                        save_dataset(data, force=True)
                except Exception as e:
                    if debug:
                        if stop_debug:
                            debug = False

                        import pdb
                        pdb.post_mortem(e.__traceback__)
                    else:
                        print('Encountered Error {}. Skipping...'.format(e))

        attach_extra_dataset_columns(dataset_path)



import numpy as np
from matplotlib.path import Path

import xarray as xr
from arpes.provenance import update_provenance
from arpes.typing import DataType
from arpes.utilities import normalize_to_spectrum

__all__ = ('polys_to_mask', 'apply_mask', 'raw_poly_to_mask',
           'apply_mask_to_coords',)


def raw_poly_to_mask(poly):
    """
    There's not currently much metadata attached to masks, but this is
    around if we ever decide that we need to implement more
    complicated masking schemes.

    In particular, we might want to store also whether the interior
    or exterior is the masked region, but this is functionally achieved
    for now with the `invert` flag in other functions.

    :param poly: Polygon implementing a masked region.
    :return:
    """
    return {
        'poly': poly,
    }


def polys_to_mask(mask_dict, coords, shape, radius=None, invert=False):
    """
    Converts a mask definition in terms of the underlying polygon to a True/False
    mask array using the coordinates and shape of the target data.

    This process "specializes" a mask to a particular shape, whereas masks given by
    polygon definitions are general to any data with appropriate dimensions, because
    waypoints are given in unitful values rather than index values.
    :param mask_dict:
    :param coords:
    :param shape:
    :param radius:
    :param invert:
    :return:
    """

    dims = mask_dict['dims']
    polys = mask_dict['polys']

    polys = [[[np.searchsorted(coords[dims[i]], coord) for i, coord in enumerate(p)] for p in poly] for poly in polys]

    mask_grids = np.meshgrid(*[np.arange(s) for s in shape])
    mask_grids = tuple(k.flatten() for k in mask_grids)

    points = np.vstack(mask_grids).T

    mask = None
    for poly in polys:
        grid = Path(poly).contains_points(points, radius=radius or 0)
        grid = grid.reshape(list(shape)[::-1]).T

        if mask is None:
            mask = grid
        else:
            mask = np.logical_or(mask, grid)

    if invert:
        mask = np.logical_not(mask)

    return mask


def apply_mask_to_coords(data: xr.Dataset, mask, dims, invert=True):
    p = Path(mask['poly'])

    as_array = np.stack([data.data_vars[d].values for d in dims], axis=-1)
    shape = as_array.shape
    dest_shape = shape[:-1]
    new_shape = [np.prod(dest_shape), len(dims)]

    mask = p.contains_points(as_array.reshape(new_shape)).reshape(dest_shape)
    if invert:
        mask = np.logical_not(mask)

    return mask


@update_provenance('Apply boolean mask to data')
def apply_mask(data: DataType, mask, replace=np.nan, radius=None, invert=False):
    """
    Applies a logical mask, i.e. one given in terms of polygons, to a specific
    piece of data. This can be used to set values outside or inside a series of
    polygon masks to a given value or to NaN.

    Expanding or contracting the masked region can be accomplished with the
    radius argument, but by default strict inclusion is used.

    Some masks include a `fermi` parameter which allows for clipping the detector
    boundaries in a semi-automated fashion. If this is included, only 200meV above the Fermi
    level will be included in the returned data. This helps to prevent very large
    and undesirable regions filled with only the replacement value which can complicate
    automated analyses that rely on masking.

    :param data: Data to mask.
    :param mask: Logical definition of the mask, appropriate for passing to `polys_to_mask`
    :param replace: The value to substitute for pixels masked.
    :param radius: Radius by which to expand the masked area.
    :param invert: Allows logical inversion of the masked parts of the data. By default,
                   the area inside the polygon sequence is replaced by `replace`.
    :return:
    """
    data = normalize_to_spectrum(data)
    fermi = mask.get('fermi')

    if isinstance(mask, dict):
        dims = mask.get('dims', data.dims)
        mask = polys_to_mask(mask, data.coords, [s for i, s in enumerate(data.shape) if data.dims[i] in dims], radius=radius, invert=invert)

    masked_data = data.copy(deep=True)
    masked_data.values = masked_data.values * 1.0
    masked_data.values[mask] = replace

    if fermi is not None:
        return masked_data.sel(eV=slice(None, fermi + 0.2))

    return masked_data



import xarray as xr
from arpes.fits import GStepBModel, broadcast_model
from arpes.provenance import update_provenance
from arpes.typing import DataType
from arpes.utilities.math import shift_by
from arpes.utilities.normalize import normalize_to_spectrum

__all__ = ('apply_cycle_fermi_edge_correction', 'build_cycle_fermi_edge_correction',)


@update_provenance('Build Fermi edge correction for cycle dimension')
def build_cycle_fermi_edge_correction(data: DataType, energy_range=None):
    arr = normalize_to_spectrum(data)

    if 'pixels' in arr.dims or 'phi' in arr.dims:
        arr = arr.S.region_sel('wide_angular')

    if energy_range is None:
        energy_range = slice(-0.1, 0.1)

    arr = arr.S.sum_other(['eV', 'cycle'])
    return broadcast_model(GStepBModel, arr.sel(eV=energy_range), 'cycle')


@update_provenance('Apply Fermi edge correction for cycle dimension')
def apply_cycle_fermi_edge_correction(data: DataType, energy_range=None, shift=True):
    correction = build_cycle_fermi_edge_correction(data, energy_range)

    if shift:
        correction_values = correction.T.map(lambda x: x.params['center'].value)
        correction_shift = - correction_values / data.S.spectrum.T.stride(generic_dim_names=False)['eV']
        dataarr = data.S.spectrum
        shifted = xr.DataArray(
            shift_by(dataarr.values, correction_shift, axis=dataarr.dims.index('eV'),
                     by_axis=dataarr.dims.index('cycle'), order=1),
            dataarr.coords,
            dataarr.dims,
            attrs=dataarr.attrs.copy()
        )
        if 'id' in shifted.attrs:
            del shifted.attrs['id']

        return shifted
    else:
        raise NotImplementedError('Need to do l2 fit, then shift axis')



"""
Some utilities for optics and optical design. This and the utilities in arpes.laser
should maybe be grouped into a separate place.

We don't do any astigmatism aware calculations here, which might be of practical utility.

Things to offer in the future:

1. Calculating beam characteristics (beam divergence, focus, etc) from a sequence of
   knife edge tests or beam profiler images in order to facilitate common tasks like beam
   collimation, focusing, or preparing a new Tr-ARPES setup.
2. Nonlinear optics utilities including damage thresholds to allow simpler design of
   harmonic generation for Tr-ARPES.
"""

# pylint: disable=invalid-name,no-value-for-parameter

import numpy as np

__all__ = ('waist', 'waist_R', 'rayleigh_range', 'lens_transfer', 'magnification', 'waist_from_divergence', 'waist_from_rr')


def waist(wavelength, z, z_R):
    raise NotImplementedError('waist not implemented')


def waist_R(waist_0, m_squared=1):
    return np.sqrt(m_squared) * waist()


def waist_from_rr(wavelength, rayleigh_rng):
    return np.sqrt((wavelength * rayleigh_rng) / np.pi)


def rayleigh_range(wavelength, beam_waist, m_squared=1):
    return np.pi * (beam_waist ** 2) / (m_squared * wavelength)


def lens_transfer(s, f, beam_rayleigh_range, m_squared=1):
    """
    Produces s''
    :param s:
    :param f:
    :param beam_rayleigh_range:
    :param m_squared:
    :return:
    """

    t = 1/f - 1/(s + (beam_rayleigh_range/m_squared) ** 2/(s - f))
    return 1 / t


def waist_from_divergence(wavelength, half_angle_divergence):
    return wavelength / (np.pi * half_angle_divergence)


def magnification(s, f, beam_rayleigh_range, m_squared=1):
    """
    Calculates the magnification offered by a lens system.
    :param s:
    :param f:
    :param beam_rayleigh_range:
    :param m_squared:
    :return:
    """
    return 1/np.sqrt((1 - s/f) ** 2 + (beam_rayleigh_range/f/m_squared)**2)



from .band_analysis import *
from .decomposition import *
from .deconvolution import *
from .derivative import *
from .filters import *
from .general import *
from .mask import *
from .tarpes import *
from .pocket import *
from .gap import *
from .savitzky_golay import *
from .xps import *
from .kfermi import *



"""
The core IO facilities supported by PyARPES. The most important here are the
data loading functions (simpe_load, fallback_load, load_without_dataset, load_example_data),
pickling utilities, data stitching, and dataset manipulation functions.
"""

import json
import os.path
import pickle
import typing
import uuid

from pathlib import Path

import numpy as np
import pandas as pd
import xarray as xr

import arpes.config
from arpes.config import (CLEAVE_RECORD, CONFIG, DATASET_CACHE_PATH,
                          DATASET_CACHE_RECORD, WorkspaceManager)
from arpes.endstations import load_scan
from arpes.exceptions import ConfigurationError
from arpes.typing import DataType
from arpes.utilities import (FREEZE_PROPS, WHITELIST_KEYS, clean_xlsx_dataset,
                             unwrap_attrs_dict, unwrap_datavar_attrs,
                             wrap_datavar_attrs)

__all__ = (
    'simple_load', 'direct_load', 'fallback_load', 'load_dataset', 'save_dataset', 'delete_dataset',
    'load_without_dataset', 'load_example_data',
    'save_dataset_for_export',
    'dataset_exists', 'is_a_dataset', 'load_dataset_attrs', 'easy_pickle',
    'sld', 'dld', 'stitch', 'fld',
)


def load_without_dataset(file: typing.Union[str, Path], location=None, **kwargs):
    file = str(Path(file).absolute())

    if location is None:
        raise ValueError('You must provide a location indicating the endstation or instrument used directly when '
                         'loading data without a dataset.')

    return load_scan(dict(file=file, location=location), **kwargs)


def load_example_data() -> xr.Dataset:
    file = Path(__file__).parent / 'example_data' / 'main_chamber_cut_0.fits'
    return load_without_dataset(file=file, location='ALG-MC')


def stitch(df_or_list, attr_or_axis, built_axis_name=None, sort=True):
    """
    Stitches together a sequence of scans or a DataFrame in order to provide a unified dataset along a specified axis

    :param df_or_list: list of the files to load
    :param attr_or_axis: coordinate or attribute in order to promote to an index. I.e. if 't_a' is specified,
    we will create a new axis corresponding to the temperature and concatenate the data along this axis
    :return:
    """

    list_of_files = None
    if isinstance(df_or_list, (pd.DataFrame,)):
        list_of_files = list(df_or_list.index)
    else:
        if not isinstance(df_or_list, (list, tuple,)):
            raise TypeError('Expected an interable for a list of the scans to stitch together')

        list_of_files = list(df_or_list)

    if built_axis_name is None:
        built_axis_name = attr_or_axis

    if not list_of_files:
        raise ValueError('Must supply at least one file to stitch')

    loaded = [f if isinstance(f, (xr.DataArray, xr.Dataset)) else simple_load(f)
              for f in list_of_files]

    for i, loaded_file in enumerate(loaded):
        value = None
        if isinstance(attr_or_axis, (list, tuple)):
            value = attr_or_axis[i]
        elif attr_or_axis in loaded_file.attrs:
            value = loaded_file.attrs[attr_or_axis]
        elif attr_or_axis in loaded_file.coords:
            value = loaded_file.coords[attr_or_axis]

        loaded_file.coords[built_axis_name] = value

    if sort:
        loaded.sort(key=lambda x: x.coords[built_axis_name])

    concatenated = xr.concat(loaded, dim=built_axis_name)
    if 'id' in concatenated.attrs:
        del concatenated.attrs['id']

    from arpes.provenance import provenance_multiple_parents
    provenance_multiple_parents(concatenated, loaded, {
        'what': 'Stitched together separate datasets',
        'by': 'stitch',
        'dim': built_axis_name,
    })

    return concatenated


def file_for_pickle(name):
    path = Path('picklejar', '{}.pickle'.format(name))
    path.parent.mkdir(exist_ok=True)
    return str(path)


def load_pickle(name):
    with open(file_for_pickle(name), 'rb') as file:
        return pickle.load(file)


def save_pickle(data, name):
    pickle.dump(data, open(file_for_pickle(name), 'wb'))


def easy_pickle(data_or_str, name=None):
    if isinstance(data_or_str, str) or name is None:
        return load_pickle(data_or_str)

    assert isinstance(name, str)
    save_pickle(data_or_str, name)


def _id_for(data):
    if isinstance(data, (xr.DataArray, xr.Dataset)):
        data = data.attrs['id']

    if isinstance(data, pd.Series):
        data = data.id

    if isinstance(data, uuid.UUID):
        data = str(data)

    return data


def _filename_for(data):
    return os.path.join(DATASET_CACHE_PATH, _id_for(data) + '.nc')


def _filename_for_attrs(data):
    return os.path.join(DATASET_CACHE_PATH, _id_for(data) + '.json')


def delete_dataset(arr_or_uuid):
    if isinstance(arr_or_uuid, xr.DataArray):
        return delete_dataset(arr_or_uuid.attrs['id'])

    assert isinstance(arr_or_uuid, str)

    fname = _filename_for(arr_or_uuid)
    if os.path.exists(fname):
        os.remove(fname)


def save_dataset(arr: DataType, filename=None, force=False):
    """
    Persists a dataset to disk. In order to serialize some attributes, you may need to modify wrap and unwrap arrs above
    in order to make sure a parameter is saved.

    In some cases, such as when you would like to add information to the attributes,
    it is nice to be able to force a write, since a write would not take place if the file is already on disk.
    To do this you can set the ``force`` attribute.
    :param arr:
    :param force:
    :return:
    """
    import arpes.xarray_extensions # pylint: disable=unused-import, redefined-outer-name

    with open(DATASET_CACHE_RECORD, 'r') as cache_record:
        records = json.load(cache_record)

    filename = filename or _filename_for(arr)
    if filename is None:
        filename = _filename_for(arr)
        attrs_filename = _filename_for_attrs(arr)
    else:
        attrs_filename = filename + '.attrs.json'

    if 'id' in arr.attrs and arr.attrs['id'] in records:
        if force:
            if os.path.exists(filename):
                os.replace(filename, filename + '.keep')
            if os.path.exists(attrs_filename):
                os.replace(attrs_filename, attrs_filename + '.keep')
        else:
            return

    df = arr.attrs.pop('df', None)
    arr.attrs.pop('', None)  # protect against totally stripped attribute names
    arr = wrap_datavar_attrs(arr, original_data=arr)
    ref_attrs = arr.attrs.pop('ref_attrs', None)

    arr.to_netcdf(filename, engine='netcdf4')
    with open(attrs_filename, 'w') as file:
        json.dump(arr.attrs, file)

    if 'id' in arr.attrs:
        first_write = arr.attrs['id'] not in records
        records[arr.attrs['id']] = {
            'file': filename,
            **{k: v for k, v in arr.attrs.items() if k in WHITELIST_KEYS}
        }
    else:
        first_write = False

    # this was a first write
    if first_write:
        with open(DATASET_CACHE_RECORD, 'w') as cache_record:
            json.dump(records, cache_record, sort_keys=True, indent=2)

    if ref_attrs is not None:
        arr.attrs['ref_attrs'] = ref_attrs

    arr = unwrap_datavar_attrs(arr)
    if df is not None:
        arr.attrs['df'] = df


def save_dataset_for_export(arr: DataType, index, **kwargs):
    filename = Path('./export/{}.nc'.format(index))
    filename.parent.mkdir(exist_ok=True)

    save_dataset(arr, str(filename), **kwargs)


def is_a_dataset(dataset):
    if isinstance(dataset, (xr.Dataset, xr.DataArray)):
        return True

    if isinstance(dataset, str):
        try:
            _ = uuid.UUID(dataset)
            return True
        except ValueError:
            return False

    return False


def dataset_exists(dataset):
    if isinstance(dataset, (xr.Dataset, xr.DataArray)):
        return True

    if isinstance(dataset, str):
        filename = _filename_for(dataset)
        return os.path.exists(filename)

    return False


def load_dataset_attrs(dataset_uuid):
    filename = _filename_for_attrs(dataset_uuid)
    if not os.path.exists(filename):
        try:
            dataset = load_dataset(dataset_uuid)
            save_dataset(dataset, force=True)
        except ValueError as e:
            raise ConfigurationError(
                'Could not load attributes for {}'.format(dataset_uuid)) from e

    with open(filename, 'r') as file:
        attrs = json.load(file)
        return unwrap_attrs_dict(attrs)


def simple_load(fragment, df: pd.DataFrame = None, workspace=None, basic_prep=True):
    with WorkspaceManager(workspace):
        if df is None:
            from arpes.utilities import default_dataset  # break circular dependency
            df = default_dataset()

        def resolve_fragment(filename):
            return str(filename).split('_')[-1]

        # find a soft match
        files = df.index

        def strip_left_zeros(value):
            if len(value) == 1:
                return value

            return value.lstrip('0')

        if isinstance(fragment, (int, np.int32, np.int64,)):
            numbers = [int(f) for f in [strip_left_zeros(''.join(c for c in resolve_fragment(f) if c.isdigit()))
                                        for f in files] if len(f)]
            index = numbers.index(fragment)
        else:
            fragment = str(fragment)
            matches = [i for i, f in enumerate(files) if fragment in f]
            if not matches:
                raise ValueError('No match found for {}'.format(fragment))
            if len(matches) > 1:
                raise ValueError('Unique match not found for {}. Options are: {}'.format(
                    fragment, [files[i] for i in matches]))
            index = matches[0]

        data = load_dataset(dataset_uuid=df.loc[df.index[index]], df=df)

        if basic_prep:
            if 'cycle' in data.indexes and len(data.coords['cycle']) == 1:
                data = data.sum('cycle', keep_attrs=True)

        return data


def direct_load(fragment, df: pd.DataFrame = None, workspace=None, file=None, basic_prep=True, **kwargs):
    """
    Loads a dataset directly, in the same manner that prepare_raw_files does, from the denormalized source format.
    This is useful for testing data loading procedures, and for quickly opening data at beamlines.

    The structure of this is very similar to simple_load, and could be shared. The only differences are in selecting
    the DataFrame with all the files at the beginning, and finally loading the data at the end.
    :param fragment:
    :param df:
    :param workspace:
    :param file:
    :param basic_prep:
    :return:
    """

    with WorkspaceManager(workspace):
        # first get our hands on a dataframe that has a list of all the files, where to find them on disk, and metadata
        if df is None:
            arpes.config.attempt_determine_workspace(lazy=True)
            if file is None:
                from arpes.utilities import default_dataset  # break circular dependency
                df = default_dataset(with_inferred_cols=False)
            else:
                if not os.path.isabs(file):
                    file = os.path.join(CONFIG['WORKSPACE']['path'], file)

                df = clean_xlsx_dataset(file, with_inferred_cols=False, write=False)

        def resolve_fragment(filename: int) -> str:
            return str(filename).split('_')[-1]

        # find a soft match
        files = df.index

        def strip_left_zeros(value: str) -> str:
            if len(value) == 1:
                return value

            return value.lstrip('0')

        if isinstance(fragment, (int, np.int32, np.int64,)):
            numbers = [int(f) for f in [strip_left_zeros(''.join(c for c in resolve_fragment(f) if c.isdigit()))
                                        for f in files] if len(f)]
            index = numbers.index(fragment)
        else:
            fragment = str(fragment)
            matches = [i for i, f in enumerate(files) if fragment in f]
            if not matches:
                raise ValueError('No match found for {}'.format(fragment))
            if len(matches) > 1:
                raise ValueError('Unique match not found for {}. Options are: {}'.format(
                    fragment, [files[i] for i in matches]))
            index = matches[0]

        scan = df.loc[df.index[index]]
        data = load_scan(dict(scan), **kwargs)

        if basic_prep:
            if 'cycle' in data.indexes and len(data.coords['cycle']) == 1:
                data = data.sum('cycle', keep_attrs=True)

        return data


sld = simple_load
dld = direct_load


def fallback_load(*args, **kwargs):
    try:
        return sld(*args, **kwargs)
    except: # pylint: disable=bare-except
        return dld(*args, **kwargs)


fld = fallback_load


def load_dataset(dataset_uuid=None, filename=None, df: pd.DataFrame = None):
    """
    You might want to prefer ``simple_load`` over calling this directly as it is more convenient.

    :param dataset_uuid: UUID of dataset to load, typically you get this from ds.loc['...'].id. This actually also
    accepts a dataframe slice so ds.loc['...'] also works.
    :param df: dataframe to use to lookup the data in. If none is provided, the result of default_dataset is used.
    :return:
    """
    if df is None:
        try:
            from arpes.utilities import default_dataset  # break circular dependency
            df = default_dataset()
        except Exception: # pylint: disable=broad-except
            pass

    if filename is None:
        filename = _filename_for(dataset_uuid)

    if not os.path.exists(filename):
        raise ValueError('%s is not cached on the FS. Did you run `prepare_raw_data`?')

    try:
        arr = xr.open_dataset(filename)
    except ValueError:
        arr = xr.open_dataarray(filename)
    arr = unwrap_datavar_attrs(arr)

    # If the sample is associated with a cleave, attach the information from that cleave
    if 'sample' in arr.attrs and 'cleave' in arr.attrs:
        full_cleave_name = '%s-%s' % (arr.attrs['sample'], arr.attrs['cleave'])

        with open(CLEAVE_RECORD, 'r') as file:
            cleaves = json.load(file)

        skip_keys = {'included_scans', 'note'}
        for k, v in cleaves.get(full_cleave_name, {}).items():
            if k not in skip_keys and k not in arr.attrs:
                arr.attrs[k] = v

    if 'ref_id' in arr.attrs:
        arr.attrs['ref_attrs'] = load_dataset_attrs(arr.attrs['ref_id'])

    for prop in FREEZE_PROPS:
        if prop not in arr.attrs:
            arr.attrs[prop] = getattr(arr.S, prop)

    arr.attrs['df'] = df

    return arr


def available_datasets(**filters):
    with open(DATASET_CACHE_RECORD, 'r') as cache_record:
        records = json.load(cache_record)

    for f, value in filters.items():
        records = {k: v for k, v in records.items() if f in v and v[f] == value}

    return records


def flush_cache(ids, delete=True):
    with open(DATASET_CACHE_RECORD, 'r') as cache_record:
        records = json.load(cache_record)

    for r_id in ids:
        if r_id in records:
            del records[r_id]

        filename = os.path.join(DATASET_CACHE_PATH, r_id + '.nc')
        if os.path.exists(filename) and delete:
            os.remove(filename)

    with open(DATASET_CACHE_RECORD, 'w') as cache_record:
        json.dump(records, cache_record, sort_keys=True, indent=2)



import functools
import time
from collections import defaultdict

from numpy import ndarray

import xarray as xr
from arpes.typing import DataType
from typing import Any, Callable, Dict, Iterator, Optional, Tuple

__all__ = ['Debounce', 'lift_dataarray_to_generic', 'iter_leaves']


def collect_leaves(tree: Dict[str, Any], is_leaf: Optional[Any] = None) -> Dict:
    """
    Produces a flat representation of the leaves, leaves with the same key are
    collected into a list in the order of appearance, but this depends on the
    dictionary iteration order.

    Example:
    collect_leaves({'a': 1, 'b': 2, 'c': {'a': 3, 'b': 4}}) -> {'a': [1, 3], 'b': [2, 4]}

    :param tree:
    :param is_leaf:
    :return:
    """
    def reducer(dd: Dict, item: Tuple[str, ndarray]) -> Dict:
        dd[item[0]].append(item[1])
        return dd

    return functools.reduce(reducer, iter_leaves(tree, is_leaf), defaultdict(list))


def iter_leaves(tree: Dict[str, Any], is_leaf: Optional[Callable] = None) -> Iterator[Tuple[str, ndarray]]:
    """
    Iterates across the leaves of a nested dictionary. Whether a particular piece
    of data counts as a leaf is controlled by the predicate `is_leaf`. By default,
    all nested dictionaries are considered not leaves, i.e. an item is a leaf if and
    only if it is not a dictionary.

    Iterated items are returned as key value pairs.

    As an example, you can easily flatten a nested structure with
    `dict(leaves(data))`

    :param tree:
    :param is_leaf:
    :return:
    """
    if is_leaf is None:
        is_leaf = lambda x: not isinstance(x, dict)

    for k, v in tree.items():
        if is_leaf(v):
            yield k, v
        else:
            for item in iter_leaves(v):
                yield item


def lift_dataarray_to_generic(f):
    """
    A functorial decorator that lifts functions with the signature

    (xr.DataArray, *args, **kwargs) -> xr.DataArray

    to one with signature

    A = typing.Union[xr.DataArray, xr.Dataset]
    (A, *args, **kwargs) -> A

    i.e. one that will operate either over xr.DataArrays or xr.Datasets.
    :param f:
    :return:
    """
    @functools.wraps(f)
    def func_wrapper(data: DataType, *args, **kwargs):
        if isinstance(data, xr.DataArray):
            return f(data, *args, **kwargs)
        else:
            assert isinstance(data, xr.Dataset)
            new_vars = {
                datavar: f(data[datavar], *args, **kwargs) for datavar in data.data_vars
            }

            for var_name, var in new_vars.items():
                if isinstance(var, xr.DataArray) and var.name is None:
                    var.name = var_name

            merged = xr.merge(new_vars.values())
            return merged.assign_attrs(data.attrs)

    return func_wrapper


class Debounce:
    def __init__(self, period):
        self.period = period  # never call the wrapped function more often than this (in seconds)
        self.count = 0  # how many times have we successfully called the function
        self.count_rejected = 0  # how many times have we rejected the call
        self.last = None  # the last time it was called

    # force a reset of the timer, aka the next call will always work
    def reset(self):
        self.last = None

    def __call__(self, f):
        @functools.wraps(f)
        def wrapped(*args, **kwargs):
            now = time.time()
            willcall = False
            if self.last is not None:
                # amount of time since last call
                delta = now - self.last
                if delta >= self.period:
                    willcall = True
                else:
                    willcall = False
            else:
                willcall = True  # function has never been called before

            if willcall:
                # set these first incase we throw an exception
                self.last = now  # don't use time.time()
                self.count += 1
                f(*args, **kwargs)  # call wrapped function
            else:
                self.count_rejected += 1
        return wrapped



"""
Store experiment level configuration here, this module also provides functions
for loading configuration in via external files, to allow better modularity
between different projects.
"""

# pylint: disable=global-statement


import json
import logging
import os.path
import pint

from pathlib import Path
from typing import Any, Optional

from arpes.exceptions import ConfigurationError

ureg = pint.UnitRegistry()

# ARPES_ROOT SHOULD BE PROVIDED THROUGH ENVIRONMENT VARIABLES, or via `setup`
DATA_PATH = None
SOURCE_ROOT = str(Path(__file__).parent)

SETTINGS = {
    'interactive': {
        'main_width': 350,
        'marginal_width': 150,
        'palette': 'magma',
    },
    'xarray_repr_mod': False,
}

# these are all set by ``update_configuration``
FIGURE_PATH = None
DATASET_PATH = None

DATASET_CACHE_PATH = None
DATASET_CACHE_RECORD = None # .json file that holds normalized files

# .json file that records which files are linked to the same physical sample, currently unused
CLEAVE_RECORD = None

PIPELINE_SHELF = None
PIPELINE_JSON_SHELF = None


def generate_cache_files() -> None:
    global DATASET_CACHE_RECORD
    global CLEAVE_RECORD

    for record_file in [DATASET_CACHE_RECORD, CLEAVE_RECORD]:
        if not Path(record_file).exists():
            with open(record_file, 'w') as f:
                json.dump({}, f)


def update_configuration(user_path: Optional[str] = None) -> None:
    global FIGURE_PATH
    global DATASET_PATH
    global DATASET_CACHE_PATH
    global DATASET_CACHE_RECORD
    global CLEAVE_RECORD
    global PIPELINE_SHELF
    global PIPELINE_JSON_SHELF

    try:
        FIGURE_PATH = os.path.join(user_path, 'figures')
        DATASET_PATH = os.path.join(user_path, 'datasets')

        DATASET_CACHE_PATH = os.path.join(user_path, 'cache')

        DATASET_CACHE_RECORD = os.path.join(user_path, 'datasets', 'cache.json')
        CLEAVE_RECORD = os.path.join(user_path, 'datasets', 'cleaves.json')

        PIPELINE_SHELF = os.path.join(user_path, 'datasets', 'pipeline.shelf')
        PIPELINE_JSON_SHELF = os.path.join(user_path, 'datasets', 'pipeline.shelf.json')

        generate_cache_files()
    except TypeError:
        pass


CONFIG = {
    'WORKSPACE': {},
    'CURRENT_CONTEXT': None,
}


class WorkspaceManager:
    def __init__(self, workspace: Optional[Any] = None) -> None:
        self._cached_workspace = None
        self._workspace = workspace

    def __enter__(self) -> None:
        global CONFIG
        self._cached_workspace = CONFIG['WORKSPACE']

        if not self._workspace:
            return

        if not CONFIG['WORKSPACE']:
            attempt_determine_workspace()

        workspace_path = Path(CONFIG['WORKSPACE']['path']).parent / self._workspace

        if workspace_path.exists():
            CONFIG['WORKSPACE'] = dict(CONFIG['WORKSPACE'])
            CONFIG['WORKSPACE']['name'] = self._workspace
            CONFIG['WORKSPACE']['path'] = str(workspace_path)
        else:
            raise ValueError('Could not find workspace: {}'.format(self._workspace))

    def __exit__(self, *args: Any) -> None:
        global CONFIG
        CONFIG['WORKSPACE'] = self._cached_workspace


def workspace_name_is_valid(workspace_name):
    return workspace_name in os.listdir(DATA_PATH)


def workspace_matches(path):
    files = os.listdir(path)
    acceptable_suffixes = {'.xlx', '.xlsx', '.numbers'}

    return 'data' in files and any(Path(f).suffix in acceptable_suffixes for f in files)


def attempt_determine_workspace(value=None, permissive=False, lazy=False, current_path=None):
    # first search upwards from the current directory at most three folders:
    if lazy and not CONFIG['WORKSPACE']:
        return

    try:
        current_path = os.getcwd()
        for _ in range(3):
            print('Checking: {}'.format(current_path))
            if workspace_matches(current_path):
                CONFIG['WORKSPACE'] = {
                    'path': current_path,
                    'name': Path(current_path).name
                }
                return

            current_path = Path(current_path).parent
    except Exception: # pylint: disable=broad-except
        pass

    if not CONFIG['WORKSPACE']:
        if current_path is None:
            current_path = os.path.realpath(os.getcwd())

        option = None
        skip_dirs = {'experiments', 'experiment', 'exp', 'projects', 'project'}

        if os.path.realpath(DATASET_PATH) in str(current_path):
            path_fragment = current_path.split(os.path.realpath(DATASET_PATH))[1]
            option = [x for x in path_fragment.split('/') if len(x) and x not in skip_dirs][0]
            # we are in a dataset, we can use the folder name in order to configure

        if value is not None:
            option = value

        if workspace_name_is_valid(option):
            logging.warning('Automatically inferring that the workspace is "{}"'.format(option))
            CONFIG['WORKSPACE'] = option

    if not CONFIG['WORKSPACE'] and not permissive:
        raise ConfigurationError('You must provide a workspace.')


def load_json_configuration(filename):
    """
    Flat updates the configuration. Beware that this doesn't update nested data.
    I will adjust if it turns out that there is a use case for nested configuration
    """
    with open(filename) as config_file:
        CONFIG.update(json.load(config_file))


try:
    from local_config import * # pylint: disable=wildcard-import
except ImportError:
    logging.warning("Could not find local configuration file. If you don't "
                    "have one, you can safely ignore this message.")


def override_settings(new_settings):
    from arpes.utilities.collections import deep_update

    global SETTINGS
    deep_update(SETTINGS, new_settings)


# try to generate cache files if they do not exist
try:
    generate_cache_files()
except TypeError:
    pass


# load plugins
def load_plugins() -> None:
    import arpes.endstations.plugin as plugin
    from arpes.endstations import add_endstation
    import importlib

    skip_modules = {'__pycache__', '__init__'}
    plugins_dir = str(Path(plugin.__file__).parent)
    modules = os.listdir(plugins_dir)
    modules = [m if os.path.isdir(os.path.join(plugins_dir, m))
               else os.path.splitext(m)[0] for m in modules if m not in skip_modules]

    for module in modules:
        try:
            loaded_module = importlib.import_module('arpes.endstations.plugin.{}'.format(module))
            for item in loaded_module.__all__:
                add_endstation(getattr(loaded_module, item))
        except (AttributeError, ImportError):
            pass


def use_tex(rc_text_should_use=False):
    import matplotlib
    matplotlib.rcParams['text.usetex'] = rc_text_should_use


update_configuration()
load_plugins()



"""
Contains calibrations and information for spectrometer resolution.
"""
import math

import numpy as np

# all resolutions are given by (photon energy, entrance slit, exit slit size)
from arpes.constants import K_BOLTZMANN_MEV_KELVIN
from arpes.typing import DataType
from arpes.utilities import normalize_to_spectrum

__all__ = ('total_resolution_estimate',)

# all analyzer dimensions are given in millimeters for convenience as this
# is how slit sizes are typically reported
def r8000(slits):
    return {
        'type': 'HEMISPHERE',
        'slits': slits,
        'radius': 200,
        'angle_resolution': 0.1 * np.pi / 180,
    }

def analyzer_resolution(analyzer_information, slit_width=None, slit_number=None,
                        pass_energy=10):
    """
    Estimates analyzer resolution from slit dimensioons pass energy, and analyzer radius.
    :param analyzer_information:
    :param slit_width:
    :param slit_number:
    :param pass_energy:
    :return:
    """
    if slit_width is None:
        slit_width = analyzer_information['slits'][slit_number]

    return 1000 * pass_energy * (slit_width / (2 * analyzer_information['radius']))

SPECTROMETER_INFORMATION = {
    'BL403':  r8000([0.05, 0.1, 0.2, 0.2, 0.3, 0.3, 0.5, 0.5, 0.8])
}

MERLIN_BEAMLINE_RESOLUTION = {
    'LEG': {
        # 40 um by 40 um slits
        (25, (40, 40)): 9.5,
        (30, (40, 40)): 13.5,
        (35, (40, 40)): 22.4,
    },
    'HEG': {
        # 30 um by 30 um slits
        (30, (30, 30)): 5.2,
        (35, (30, 30)): 5.4,
        (40, (30, 30)): 5.0,
        (45, (30, 30)): 5.5,
        (50, (30, 30)): 5.4,
        (55, (30, 30)): 5.5,
        (60, (30, 30)): 6.4,
        (65, (30, 30)): 6.5,
        (70, (30, 30)): 7.7,
        (75, (30, 30)): 9,
        (80, (30, 30)): 8.6,
        (90, (30, 30)): 12.4,
        (100, (30, 30)): 21.6,
        (110, (30, 30)): 35.4,

        # 50um by 50um
        (60, (50, 50)): 8.2,
        (70, (50, 50)): 10.2,
        (80, (50, 50)): 12.6,
        (90, (50, 50)): 16.5,

        # 60um by 80um
        (60, (60, 80)): 9.2,
        (70, (60, 80)): 13.0,
        (80, (60, 80)): 16.5,
        (90, (60, 80)): 22.0,

        # 90um by 140um
        (30, (90, 140)): 7.3,
        (40, (90, 140)): 9,
        (50, (90, 140)): 11.9,
        (60, (90, 140)): 16.2,
        (70, (90, 140)): 21.4,
        (80, (90, 140)): 25.8,
        (90, (90, 140)): 34,
        (100, (90, 140)): 45,
        (110, (90, 140)): 60,
    },
    # second order from the high density grating
    'HEG-2': {
        # 30um by 30um
        (90, (30, 30)): 8,
        (100, (30, 30)): 9,
        (110, (30, 30)): 9.6,
        (120, (30, 30)): 9.6,
        (130, (30, 30)): 12,
        (140, (30, 30)): 15,
        (150, (30, 30)): 13,

        # 50um by 50um
        (90, (50, 50)): 10.3,
        (100, (50, 50)): 10.5,
        (110, (50, 50)): 13.2,
        (120, (50, 50)): 14,
        (130, (50, 50)): 19,
        (140, (50, 50)): 22,
        (150, (50, 50)): 22,

        # 60um by 80um
        (90, (60, 80)): 12.8,
        (100, (60, 80)): 15,
        (110, (60, 80)): 16.4,
        (120, (60, 80)): 19,
        (130, (60, 80)): 27,
        (140, (60, 80)): 30,

        # 90um by 140um
        (90, (90, 140)): 19,
        (100, (90, 140)): 21,
        (110, (90, 140)): 28,
        (120, (90, 140)): 31,
        (130, (90, 140)): 37,
        (140, (90, 140)): 41,
        (150, (90, 140)): 49,
    },
}

ENDSTATIONS_BEAMLINE_RESOLUTION = {
    'BL403': MERLIN_BEAMLINE_RESOLUTION,
}


def analyzer_resolution_estimate(data: DataType, meV=False):
    """
    For hemispherical analyzers, this can be determined by the slit
    and pass energy settings.

    Roughly,
    :param data:
    :return:
    """
    data = normalize_to_spectrum(data)

    endstation = data.S.endstation
    spectrometer_info = SPECTROMETER_INFORMATION[endstation]

    spectrometer_settings = data.S.spectrometer_settings

    return analyzer_resolution(spectrometer_info, slit_number=spectrometer_settings['slit'],
                               pass_energy=spectrometer_settings['pass_energy']) * (1 if meV else 0.001)


def energy_resolution_from_beamline_slit(table, photon_energy, exit_slit_size):
    """
    Assumes an exact match on the photon energy, though that interpolation
    could also be pulled into here...
    :param table:
    :param photon_energy:
    :param exit_slit_size:
    :return:
    """

    by_slits = {k[1]: v for k, v in table.items() if k[0] == photon_energy}
    if exit_slit_size in by_slits:
        return by_slits[exit_slit_size]

    slit_area = exit_slit_size[0] * exit_slit_size[1]
    by_area = {int(k[0] * k[1]): v for k, v in by_slits.items()}

    if len(by_area) == 1:
        return list(by_area.values())[0] * slit_area / (list(by_area.keys())[0])

    try:
        low = max(k for k in by_area.keys() if k <= slit_area)
        high = min(k for k in by_area.keys() if k >= slit_area)
    except ValueError:
        if slit_area > max(by_area.keys()):
            # use the largest and second largest
            high = max(by_area.keys())
            low = max(k for k in by_area.keys() if k < high)
        else:
            # use the smallest and second smallest
            low = min(by_area.keys())
            high = min(k for k in by_area.keys() if k > low)

    return by_area[low] + (by_area[high] - by_area[low]) * (slit_area - low) / (high - low)


def beamline_resolution_estimate(data: DataType, meV=False):
    data = normalize_to_spectrum(data)
    resolution_table = ENDSTATIONS_BEAMLINE_RESOLUTION[data.S.endstation]

    if isinstance(list(resolution_table.keys())[0], str):
        # need grating information
        settings = data.S.beamline_settings
        resolution_table = resolution_table[settings['grating']]

        all_keys = list(resolution_table.keys())
        hvs = set(k[0] for k in all_keys)

        low_hv = max(hv for hv in hvs if hv < settings['hv'])
        high_hv = min(hv for hv in hvs if hv >= settings['hv'])

        slit_size = (settings['entrance_slit'], settings['exit_slit'],)
        low_hv_res = energy_resolution_from_beamline_slit(
            resolution_table, low_hv, slit_size)
        high_hv_res = energy_resolution_from_beamline_slit(
            resolution_table, high_hv, slit_size)

        # interpolate between nearest values
        return low_hv_res + (high_hv_res - low_hv_res) * \
                            (settings['hv'] - low_hv) / (high_hv - low_hv) * (1000 if meV else 1)

    raise NotImplementedError()


def thermal_broadening_estimate(data: DataType, meV=False):
    """
    Simple Fermi-Dirac broadening
    :param data:
    :return:
    """
    return normalize_to_spectrum(data).S.temp * K_BOLTZMANN_MEV_KELVIN * (1 if meV else 0.001)


def total_resolution_estimate(data: DataType, include_thermal_broadening=False, meV=False):
    """
    Gives the quadrature sum estimate of the resolution of an ARPES
    spectrum that is decorated with appropriate information.

    For synchrotron ARPES, this typically means the scan has the photon energy,
    exit slit information and analyzer slit settings
    :return:
    """

    thermal_broadening = 0
    if include_thermal_broadening:
        thermal_broadening = thermal_broadening_estimate(data, meV=meV)
    return math.sqrt(
        beamline_resolution_estimate(data, meV=meV) ** 2 +
        analyzer_resolution_estimate(data, meV=meV) ** 2 +
        thermal_broadening ** 2
    )



import matplotlib.colors
import matplotlib.pyplot as plt
import numpy as np

import xarray as xr
from arpes.plotting.utils import imshow_arr, path_for_plot
from arpes.provenance import save_plot_provenance
from arpes.utilities import normalize_to_spectrum


@save_plot_provenance
def false_color_plot(data_r: xr.Dataset, data_g: xr.Dataset, data_b: xr.Dataset, ax=None, out=None, invert=False, pmin=0, pmax=1, **kwargs):
    data_r, data_g, data_b = [normalize_to_spectrum(d) for d in (data_r, data_g, data_b)]
    fig = None
    if ax is None:
        fig, ax = plt.subplots(figsize=kwargs.get('figsize', (7, 5,)))

    def normalize_channel(channel):
        channel -= np.percentile(channel, 100 * pmin)
        channel[channel > np.percentile(channel, 100 * pmax)] = np.percentile(channel, 100 * pmax)
        channel = channel / np.max(channel)
        return channel

    cs = dict(data_r.coords)
    cs['dim_color'] = [1, 2, 3]

    arr = xr.DataArray(
        np.stack([normalize_channel(data_r.values),
                  normalize_channel(data_g.values),
                  normalize_channel(data_b.values)], axis=-1),
        coords=cs,
        dims=list(data_r.dims) + ['dim_color'],
    )

    if invert:
        vs = arr.values
        vs[vs > 1] = 1
        hsv = matplotlib.colors.rgb_to_hsv(vs)
        hsv[:,:,2] = 1 - hsv[:,:,2]
        arr.values = matplotlib.colors.hsv_to_rgb(hsv)

    imshow_arr(arr, ax=ax)

    if out is not None:
        plt.savefig(path_for_plot(out), dpi=400)
        return path_for_plot(out)

    return fig, ax



"""
Although it is not preferred, we support building data analysis pipelines.

As an analogy, this is kind of the "dual" in the sense of vector space duality:
rather than starting with data and doing things with it to perform an analysis,
you can specify how to chain and apply operations to the analysis functions and build
a pipeline or sequence of operations that can in the end be applied to data.

This has some distinct advantages:

1. You can cache computations before expensive steps, and restart calculations part way through.
   In fact, this is already supported in PyARPES. For instance, you can specify a pipline that does
       i.   An expensive correction
       ii.  Momentum space conversion
       iii. An expensive analysis in momentum space
   If you run data through this pipeline and have to stop during step ii., the next time
   you run the pipeline, it will start with the cached result from step i. instead of recomputing
   this value.
2. Systematizing certain kinds of analysis

The core of this is `compose` which takes two pipelines (including atomic elements
like single functions) and returns their composition, which can be paused and restarted between
the two parts. Atomic elements can be constructed with `pipeline`.

In practice though, much of ARPES analysis occurs at scales too small to make this useful,
and interactivity tends to be much preferred to rigidity. PyARPES nevertheless offers this as an
option, as well as trying to provide support for reproducible and understandable scientific
analyses without sacrificing interativity and a tight feedback loop for the experimenter.
"""

import json

from typing import Union

import xarray as xr

import arpes.config
import arpes.io


def normalize_data(data: Union[xr.DataArray, xr.Dataset, str]):
    if isinstance(data, xr.DataArray):
        assert 'id' in data.attrs
        return data.attrs['id']

    if isinstance(data, xr.Dataset):
        raise TypeError('xarray.Dataset is not supported as a normalizable dataset')

    assert isinstance(data, str)

    return data


def denormalize_data(data):
    if isinstance(data, str):
        # Try to parse it as a UUID, load or retrieve appropriate dataset
        pass

    return data


def computation_hash(pipeline_name, data, intern_kwargs, *args, **kwargs):
    return json.dumps({
        'pipeline_name': pipeline_name,
        'data': normalize_data(data),
        'args': args,
        'kwargs': {k: v for k, v in kwargs.items() if k in intern_kwargs},
    }, sort_keys=True)


def cache_computation(key, data):
    try:
        if isinstance(data, xr.DataArray):
            # intern the computation
            arpes.io.save_dataset(data)
            data = normalize_data(data)

        return data
    except Exception as e:
        raise e


class PipelineRollbackException(Exception):
    pass


def pipeline(pipeline_name=None, intern_kwargs=None):
    if intern_kwargs is None:
        intern_kwargs = set()

    # TODO write simple tests for the pipeline flags to ensure correct caching conditions
    def pipeline_decorator(f):
        def func_wrapper(data, flush=False, force=False, debug=False, verbose=True, *args, **kwargs):
            key = computation_hash(pipeline_name or f.__name__,
                                   data, intern_kwargs, *args, **kwargs)
            if debug:
                print(pipeline_name or f.__name__, key)

            def echo(v):
                if verbose:
                    print('{}: {}'.format(pipeline_name or f.__name__, v))

            with open(arpes.config.PIPELINE_JSON_SHELF, 'r') as file:
                # Currently we are using JSON because of a bug in the implementation
                # of python's shelf that causes it to hang forever.
                # A better long term solution here is to use a database or KV-store
                records = json.load(file)

            if key in records and not force:
                if (not arpes.io.is_a_dataset(key) or arpes.io.dataset_exists(key)):
                    value = records[key]
                    if flush:
                        # remove the record of the cached computation, and delete the filesystem cache
                        # for the computation
                        del records[key]
                        if arpes.io.dataset_exists(key):
                            arpes.io.delete_dataset(key)

                    with open(arpes.config.PIPELINE_JSON_SHELF, 'w') as file:
                        json.dump(records, file)

                    echo(value)
                    return value
                else:
                    del records[key]
                    with open(arpes.config.PIPELINE_JSON_SHELF, 'w') as file:
                        json.dump(records, file)
                    raise PipelineRollbackException()
            elif flush:
                return None

            try:
                if isinstance(data, str):
                    data = arpes.io.load_dataset(data)
            except ValueError:
                pass
            finally:
                computed = f(data, *args, **kwargs)
                records[key] = cache_computation(key, computed)
                with open(arpes.config.PIPELINE_JSON_SHELF, 'w') as file:
                    json.dump(records, file)
                echo(normalize_data(computed))

            return computed

        return func_wrapper

    return pipeline_decorator


def compose(*pipelines):
    def composed(data, *args, **kwargs):
        max_restarts = len(pipelines)
        while max_restarts:
            data_in_process = data

            try:
                for next_pipeline in pipelines:
                    data_in_process = next_pipeline(data_in_process, *args, **kwargs)

                return data_in_process
            except PipelineRollbackException as e:
                if not max_restarts:
                    raise e

                max_restarts -= 1
                continue

    return composed
